"""Unit tests for reviewboard.datagrids.columns.GroupsColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import GroupsColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class GroupsColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.GroupsColumn.

    Version Added:
        8.0
    """

    column = GroupsColumn()

    def test_render_data_with_no_groups(self) -> None:
        """Testing GroupsColumn.render_data with no groups"""
        review_request = self.create_review_request(publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_one_group(self) -> None:
        """Testing GroupsColumn.render_data with one group"""
        review_request = self.create_review_request(publish=True)
        group = self.create_review_group(name='group1')
        review_request.target_groups.add(group)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, 'group1')

    def test_render_data_with_multiple_groups(self) -> None:
        """Testing GroupsColumn.render_data with multiple groups"""
        review_request = self.create_review_request(publish=True)
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')
        review_request.target_groups.add(group1, group2)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, 'group1 group2')

    def test_to_json_with_no_groups(self) -> None:
        """Testing GroupsColumn.to_json with no groups"""
        review_request = self.create_review_request(publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            [])

    def test_to_json_with_groups(self) -> None:
        """Testing GroupsColumn.to_json with groups"""
        review_request = self.create_review_request(publish=True)
        group1 = self.create_review_group(name='group1')
        group2 = self.create_review_group(name='group2')
        review_request.target_groups.add(group1, group2)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            [group1, group2])



"""A hook for registering new integration classes."""

from __future__ import annotations

from djblets.extensions.hooks import ExtensionHookPoint
from djblets.integrations.hooks import BaseIntegrationHook

from reviewboard.integrations.base import GetIntegrationManagerMixin


class IntegrationHook(GetIntegrationManagerMixin, BaseIntegrationHook,
                      metaclass=ExtensionHookPoint):
    """A hook for registering new integration classes.

    Integrations enable Review Board to connect with third-party services in
    specialized ways. This class makes it easy to register new integrations on
    an extension, binding their lifecycles to that of the extension.
    """



"""Change WebAPIToken.extra_data and policy to allow NULL values.

Version Added:
    2.5
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('WebAPIToken', 'policy', initial=None, null=True),
    ChangeField('WebAPIToken', 'extra_data', initial=None, null=True),
]



from django.contrib.auth.models import User
from kgb import SpyAgency

from reviewboard.accounts.models import LocalSiteProfile
from reviewboard.reviews.errors import NotModifiedError
from reviewboard.reviews.models import (Group, ReviewRequest,
                                        ReviewRequestDraft)
from reviewboard.site.models import LocalSite
from reviewboard.testing import TestCase


class ReviewRequestCounterTests(SpyAgency, TestCase):
    """Unit tests for review request counters."""

    fixtures = ['test_scmtools']

    def setUp(self):
        super(ReviewRequestCounterTests, self).setUp()

        repository = self.create_repository(
            name='Test1',
            path='path1',
            tool_name='Subversion')

        self.user = User.objects.create_user(username='testuser', password='',
                                             email='user@example.com')
        self.profile = self.user.get_profile()

        self.test_site = LocalSite.objects.create(name='test')
        self.site_profile2 = \
            LocalSiteProfile.objects.create(user=self.user,
                                            profile=self.profile,
                                            local_site=self.test_site)

        self.review_request = self.create_review_request(submitter=self.user,
                                                         repository=repository)

        self.profile.star_review_request(self.review_request)

        self.site_profile = self.profile.site_profiles.get(local_site=None)
        self.assertEqual(self.site_profile.total_outgoing_request_count, 1)
        self.assertEqual(self.site_profile.pending_outgoing_request_count, 1)
        self.assertEqual(self.site_profile.starred_public_request_count, 0)

        self.group = Group.objects.create(name='test-group')
        self.group.users.add(self.user)

        self._reload_objects()
        self.assertEqual(self.site_profile2.total_outgoing_request_count, 0)
        self.assertEqual(self.site_profile2.pending_outgoing_request_count, 0)
        self.assertEqual(self.site_profile2.starred_public_request_count, 0)

    def test_new_site_profile(self):
        """Testing counters on a new LocalSiteProfile"""
        self.site_profile.delete()
        self.site_profile = \
            LocalSiteProfile.objects.create(user=self.user,
                                            profile=self.profile)
        self.assertEqual(self.site_profile.total_outgoing_request_count, 1)
        self.assertEqual(self.site_profile.pending_outgoing_request_count, 1)
        self.assertEqual(self.site_profile.starred_public_request_count, 0)

        self.review_request.publish(self.user)

        self._reload_objects()
        self.assertEqual(self.site_profile.total_outgoing_request_count, 1)
        self.assertEqual(self.site_profile.pending_outgoing_request_count, 1)
        self.assertEqual(self.site_profile.starred_public_request_count, 1)

    def test_outgoing_requests(self):
        """Testing counters with creating outgoing review requests"""
        # The review request was already created
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_people.add(self.user)
        draft.save()
        self.review_request.publish(self.user)

        self._check_counters(direct_incoming=1,
                             total_incoming=1,
                             total_outgoing=1,
                             pending_outgoing=1,
                             starred_public=1)

    def test_closing_requests(self, close_type=ReviewRequest.DISCARDED):
        """Testing counters with closing outgoing review requests"""
        # The review request was already created
        self._check_counters(total_outgoing=1, pending_outgoing=1)

        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_groups.add(self.group)
        draft.target_people.add(self.user)
        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1,
                             group_incoming=1)

        self.assertTrue(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        self.review_request.close(close_type)
        self._check_counters(total_outgoing=1)

    def test_closing_draft_requests(self, close_type=ReviewRequest.DISCARDED):
        """Testing counters with closing draft review requests"""
        # The review request was already created
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        self.assertFalse(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        self.review_request.close(close_type)
        self._check_counters(total_outgoing=1)

    def test_closing_closed_requests(self):
        """Testing counters with closing closed review requests"""
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             starred_public=1)

        self.assertTrue(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        self.review_request.close(ReviewRequest.DISCARDED)
        self._check_counters(total_outgoing=1)

        self.review_request.close(ReviewRequest.SUBMITTED)
        self._check_counters(total_outgoing=1)

    def test_closing_draft_requests_with_site(self):
        """Testing counters with closing draft review requests on LocalSite"""
        self.review_request.delete()

        self._check_counters(with_local_site=True)

        repository = self.create_repository(
            name='Test1',
            path='path1',
            tool_name='Subversion',
            local_site=self.test_site)
        self.review_request = ReviewRequest.objects.create(
            self.user,
            repository,
            local_site=self.test_site)

        self._check_counters(with_local_site=True,
                             total_outgoing=1,
                             pending_outgoing=1)

        self.assertFalse(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        self.review_request.close(ReviewRequest.DISCARDED)
        self._check_counters(with_local_site=True,
                             total_outgoing=1)

    def test_deleting_requests(self):
        """Testing counters with deleting outgoing review requests"""
        # The review request was already created
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_groups.add(self.group)
        draft.target_people.add(self.user)

        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1,
                             group_incoming=1)

        self.review_request.delete()
        self._check_counters()

    def test_deleting_draft_requests(self):
        """Testing counters with deleting draft review requests"""
        # We're simulating what a DefaultReviewer would do by populating
        # the ReviewRequest's target users and groups while not public and
        # without a draft.
        self.review_request.target_people.add(self.user)
        self.review_request.target_groups.add(self.group)

        # The review request was already created
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        self.review_request.delete()
        self._check_counters()

    def test_deleting_closed_requests(self):
        """Testing counters with deleting closed review requests"""
        # We're simulating what a DefaultReviewer would do by populating
        # the ReviewRequest's target users and groups while not public and
        # without a draft.
        self.review_request.target_people.add(self.user)
        self.review_request.target_groups.add(self.group)

        # The review request was already created
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        self.review_request.close(ReviewRequest.DISCARDED)
        self._check_counters(total_outgoing=1)

        self.review_request.delete()
        self._check_counters()

    def test_reopen_discarded_requests(self):
        """Testing counters with reopening discarded outgoing review requests
        """
        self.test_closing_requests(ReviewRequest.DISCARDED)

        self.review_request.reopen()
        self.assertFalse(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1,
                             group_incoming=1)

    def test_reopen_submitted_requests(self):
        """Testing counters with reopening submitted outgoing review requests
        """
        self.test_closing_requests(ReviewRequest.SUBMITTED)

        self.review_request.reopen()
        self.assertTrue(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1,
                             group_incoming=1)

        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1,
                             group_incoming=1)

    def test_reopen_discarded_draft_requests(self):
        """Testing counters with reopening discarded draft review requests"""
        self.assertFalse(self.review_request.public)

        self.test_closing_draft_requests(ReviewRequest.DISCARDED)

        self.review_request.reopen()
        self.assertFalse(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

    def test_reopen_submitted_draft_requests(self):
        """Testing counters with reopening submitted draft review requests"""
        self.test_closing_requests(ReviewRequest.SUBMITTED)

        # We're simulating what a DefaultReviewer would do by populating
        # the ReviewRequest's target users and groups while not public and
        # without a draft.
        self.review_request.target_people.add(self.user)
        self.review_request.target_groups.add(self.group)

        self._check_counters(total_outgoing=1)

        self.review_request.reopen()
        self.assertTrue(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1,
                             group_incoming=1)

    def test_double_publish(self):
        """Testing counters with publishing a review request twice"""
        self.assertFalse(self.review_request.public)
        self.assertEqual(self.review_request.status,
                         ReviewRequest.PENDING_REVIEW)

        # Publish the first time.
        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             starred_public=1)

        # Publish the second time.
        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             starred_public=1)

    def test_add_group(self):
        """Testing counters when adding a group reviewer"""
        draft = ReviewRequestDraft.create(self.review_request)
        draft.summary = 'Test Summary'
        draft.target_groups.add(self.group)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1,
                             group_incoming=1,
                             starred_public=1)

    def test_remove_group(self):
        """Testing counters when removing a group reviewer"""

        self.test_add_group()

        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_groups.remove(self.group)

        # There must be at least one target_group or target_people
        draft.target_people.add(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1,
                             direct_incoming=0,
                             group_incoming=1,
                             starred_public=1)

        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1)

    def test_remove_group_and_fail_publish(self):
        """Testing counters when removing a group reviewer and then
        failing to publish the draft
        """
        self.test_add_group()

        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_groups.remove(self.group)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1,
                             group_incoming=1,
                             starred_public=1)

        self.spy_on(ReviewRequestDraft.publish,
                    owner=ReviewRequestDraft,
                    call_fake=self._raise_publish_error)

        with self.assertRaises(NotModifiedError):
            self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1,
                             group_incoming=1,
                             starred_public=1)

    def test_add_person(self):
        """Testing counters when adding a person reviewer"""
        draft = ReviewRequestDraft.create(self.review_request)
        draft.summary = 'Test Summary'
        draft.target_people.add(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1)

    def test_remove_person(self):
        """Testing counters when removing a person reviewer"""
        self.test_add_person()

        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_people.remove(self.user)

        # There must be at least one target_group or target_people
        draft.target_groups.set([self.group])

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1)

        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             group_incoming=1,
                             total_incoming=1,
                             starred_public=1)

    def test_remove_person_and_fail_publish(self):
        """Testing counters when removing a person reviewer and then
        failing to publish the draft
        """
        self.test_add_person()

        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_people.remove(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1)

        self.spy_on(ReviewRequestDraft.publish,
                    owner=ReviewRequestDraft,
                    call_fake=self._raise_publish_error)

        with self.assertRaises(NotModifiedError):
            self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1)

    def test_populate_counters(self):
        """Testing counters when populated from a fresh upgrade or clear"""
        # The review request was already created
        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_groups.add(self.group)
        draft.target_people.add(self.user)
        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1,
                             direct_incoming=1,
                             starred_public=1,
                             group_incoming=1)

        LocalSiteProfile.objects.update(
            direct_incoming_request_count=None,
            total_incoming_request_count=None,
            pending_outgoing_request_count=None,
            total_outgoing_request_count=None,
            starred_public_request_count=None)
        Group.objects.update(incoming_request_count=None)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1,
                             direct_incoming=1,
                             starred_public=1,
                             group_incoming=1)

    def test_populate_counters_after_change(self):
        """Testing counter inc/dec on uninitialized counter fields"""
        # The review request was already created
        draft = ReviewRequestDraft.create(self.review_request)
        draft.target_groups.add(self.group)
        draft.target_people.add(self.user)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        LocalSiteProfile.objects.update(
            direct_incoming_request_count=None,
            total_incoming_request_count=None,
            pending_outgoing_request_count=None,
            total_outgoing_request_count=None,
            starred_public_request_count=None)
        Group.objects.update(incoming_request_count=None)

        profile_fields = [
            'direct_incoming_request_count',
            'total_incoming_request_count',
            'pending_outgoing_request_count',
            'total_outgoing_request_count',
            'starred_public_request_count',
        ]

        # Lock the fields so we don't re-initialize them on publish.
        locks = {
            self.site_profile: 1,
            self.site_profile2: 1,
        }

        for field in profile_fields:
            getattr(LocalSiteProfile, field)._locks = locks

        Group.incoming_request_count._locks = locks

        # Publish the review request. This will normally try to
        # increment/decrement the counts, which it should ignore now.
        self.review_request.publish(self.user)

        # Unlock the profiles so we can query/re-initialize them again.
        for field in profile_fields:
            getattr(LocalSiteProfile, field)._locks = {}

        Group.incoming_request_count._locks = {}

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             direct_incoming=1,
                             total_incoming=1,
                             starred_public=1,
                             group_incoming=1)

    def test_counts_with_reassignment_in_initial_draft(self):
        """Testing counters when changing review request ownership in initial
        draft
        """
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        new_user = User.objects.create_user(username='test2',
                                            password='',
                                            email='user@example.com')
        site_profile = \
            new_user.get_site_profile(self.review_request.local_site)

        draft = ReviewRequestDraft.create(self.review_request)
        draft.owner = new_user
        draft.target_people.add(draft.owner)
        draft.save()
        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=0,
                             pending_outgoing=0,
                             starred_public=1)

        site_profile = LocalSiteProfile.objects.get(pk=site_profile.pk)
        self._check_counters_on_profile(site_profile,
                                        total_outgoing=1,
                                        pending_outgoing=1,
                                        direct_incoming=1,
                                        total_incoming=1)

    def test_counts_with_reassignment_in_initial_draft_new_profile(self):
        """Testing counters when changing review request ownership in initial
        draft and new owner without initial site profile
        """
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1)

        new_user = User.objects.create_user(username='test2',
                                            password='',
                                            email='user@example.com')

        draft = ReviewRequestDraft.create(self.review_request)
        draft.owner = new_user
        draft.target_people.add(draft.owner)
        draft.save()
        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=0,
                             pending_outgoing=0,
                             starred_public=1)

        site_profile = \
            new_user.get_site_profile(self.review_request.local_site)
        self._check_counters_on_profile(site_profile,
                                        total_outgoing=1,
                                        pending_outgoing=1,
                                        direct_incoming=1,
                                        total_incoming=1)

    def test_counts_with_reassignment_after_publish(self):
        """Testing counters when changing review request ownership after
        publish
        """
        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             starred_public=1)

        new_user = User.objects.create_user(username='test2',
                                            password='',
                                            email='user@example.com')
        site_profile = \
            new_user.get_site_profile(self.review_request.local_site)

        draft = ReviewRequestDraft.create(self.review_request)
        draft.owner = new_user
        draft.target_people.add(draft.owner)
        draft.save()
        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=0,
                             pending_outgoing=0,
                             starred_public=1)

        site_profile = LocalSiteProfile.objects.get(pk=site_profile.pk)
        self._check_counters_on_profile(site_profile,
                                        total_outgoing=1,
                                        pending_outgoing=1,
                                        direct_incoming=1,
                                        total_incoming=1)

    def test_counts_with_reassignment_after_publish_new_profile(self):
        """Testing counters when changing review request ownership after
        publish and new owner without initial site profile
        """
        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             starred_public=1)

        new_user = User.objects.create_user(username='test2',
                                            password='',
                                            email='user@example.com')

        draft = ReviewRequestDraft.create(self.review_request)
        draft.owner = new_user
        draft.target_people.add(draft.owner)
        draft.save()
        self.review_request.publish(self.user)

        self._check_counters(total_outgoing=0,
                             pending_outgoing=0,
                             starred_public=1)

        site_profile = \
            new_user.get_site_profile(self.review_request.local_site)
        self._check_counters_on_profile(site_profile,
                                        total_outgoing=1,
                                        pending_outgoing=1,
                                        direct_incoming=1,
                                        total_incoming=1)

    def test_counts_with_reassignment_and_close(self):
        """Testing counters when changing review request ownership and closing
        in same operation
        """
        self.review_request.publish(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             starred_public=1)

        new_user = User.objects.create_user(username='test2',
                                            password='',
                                            email='user@example.com')
        site_profile = \
            new_user.get_site_profile(self.review_request.local_site)

        # Note that it's not normally possible to update something like an
        # owner while also closing in the same operation. Drafts don't allow
        # it. However, we have logic that considers these combinations of
        # operations, and it's technically possible to do, so we're testing
        # it here by updating the review request manually.
        self.review_request.owner = new_user
        self.review_request.status = ReviewRequest.SUBMITTED
        self.review_request.save(update_counts=True,
                                 old_submitter=self.user)

        self._check_counters(total_outgoing=0,
                             pending_outgoing=0,
                             starred_public=0)

        site_profile = LocalSiteProfile.objects.get(pk=site_profile.pk)
        self._check_counters_on_profile(site_profile,
                                        total_outgoing=1,
                                        pending_outgoing=0)

    def test_counts_with_reassignment_and_reopen(self):
        """Testing counters when changing review request ownership and
        reopening in same operation
        """
        self.review_request.close(ReviewRequest.DISCARDED)
        self.assertFalse(self.review_request.public)
        self.assertEqual(self.review_request.status, ReviewRequest.DISCARDED)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=0,
                             starred_public=0)

        new_user = User.objects.create_user(username='test2',
                                            password='',
                                            email='user@example.com')
        site_profile = \
            new_user.get_site_profile(self.review_request.local_site)

        # Note that it's not normally possible to update something like an
        # owner while also reopening in the same operation. Drafts don't allow
        # it. However, we have logic that considers these combinations of
        # operations, and it's technically possible to do, so we're testing
        # it here by updating the review request manually.
        self.review_request.owner = new_user
        self.review_request.status = ReviewRequest.PENDING_REVIEW
        self.review_request.save(update_counts=True,
                                 old_submitter=self.user)

        self._check_counters(total_outgoing=0,
                             pending_outgoing=0,
                             starred_public=0)

        site_profile = LocalSiteProfile.objects.get(pk=site_profile.pk)
        self._check_counters_on_profile(site_profile,
                                        total_outgoing=1,
                                        pending_outgoing=1)

    def test_counts_with_join_group(self):
        """Testing counters when joining a review group"""
        user2 = self.create_user()
        group2 = self.create_review_group(name='group2')

        self.create_review_request(submitter=user2,
                                   target_groups=[group2],
                                   publish=True)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=0)

        group2.users.add(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1)

    def test_counts_with_leave_group(self):
        """Testing counters when leaving a review group"""
        user2 = self.create_user()

        group2 = self.create_review_group(name='group2')
        group2.users.add(self.user)

        self.create_review_request(submitter=user2,
                                   target_groups=[group2],
                                   publish=True)

        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=1)

        group2.users.remove(self.user)
        self._check_counters(total_outgoing=1,
                             pending_outgoing=1,
                             total_incoming=0)

    def _check_counters(self, total_outgoing=0, pending_outgoing=0,
                        direct_incoming=0, total_incoming=0,
                        starred_public=0, group_incoming=0,
                        with_local_site=False):
        """Check that the counters match the expected values.

        Args:
            total_outgoing (int):
                The expected number of total outgoing review requests.

            pending_outgoing (int):
                The expected number of pending outgoing review requests.

            direct_incoming (int):
                The expected number of review requests assigned directly to the
                user.

            total_incoming (int):
                The expected number of review requests assigned either directly
                or indirectly to the user.

            starred_public (int):
                The expected number of public review requests starred by the
                user.

            group_incoming (int):
                The expected number of review requests assigned to the test
                group.

            with_local_site (bool):
                Whether to run the test for a local site.
        """
        self._reload_objects()

        if with_local_site:
            main_site_profile = self.site_profile2
            unused_site_profile = self.site_profile
        else:
            main_site_profile = self.site_profile
            unused_site_profile = self.site_profile2

        self._check_counters_on_profile(main_site_profile, total_outgoing,
                                        pending_outgoing, direct_incoming,
                                        total_incoming, starred_public)
        self.assertEqual(
            self.group.incoming_request_count,
            group_incoming,
            'Expected Group.incoming_request_count to be %s. Got %s instead.'
            % (group_incoming, self.group.incoming_request_count))

        # These should never be affected by updates on the main LocalSite we're
        # working with, so they should always be 0.
        self._check_counters_on_profile(unused_site_profile)

    def _check_counters_on_profile(self, profile, total_outgoing=0,
                                   pending_outgoing=0, direct_incoming=0,
                                   total_incoming=0, starred_public=0):
        """Check that the counters match the expected values.

        Args:
            profile (reviewboard.accounts.models.LocalSiteProfile):
                The profile object to test counts on.

            total_outgoing (int):
                The expected number of total outgoing review requests.

            pending_outgoing (int):
                The expected number of pending outgoing review requests.

            direct_incoming (int):
                The expected number of review requests assigned directly to the
                user.

            total_incoming (int):
                The expected number of review requests assigned either directly
                or indirectly to the user.

            starred_public (int):
                The expected number of public review requests starred by the
                user.
        """
        msg = 'Expected %s to be %s. Got %s instead.'

        self.assertEqual(
            profile.total_outgoing_request_count,
            total_outgoing,
            msg % ('total_outgoing_request_count', total_outgoing,
                   profile.total_outgoing_request_count))
        self.assertEqual(
            profile.pending_outgoing_request_count,
            pending_outgoing,
            msg % ('pending_outgoing_request_count', pending_outgoing,
                   profile.pending_outgoing_request_count))
        self.assertEqual(
            profile.direct_incoming_request_count,
            direct_incoming,
            msg % ('direct_incoming_request_count', direct_incoming,
                   profile.direct_incoming_request_count))
        self.assertEqual(
            profile.total_incoming_request_count,
            total_incoming,
            msg % ('total_incoming_request_count', total_incoming,
                   profile.total_incoming_request_count))
        self.assertEqual(
            profile.starred_public_request_count,
            starred_public,
            msg % ('starred_public_request_count', starred_public,
                   profile.starred_public_request_count))

    def _reload_objects(self):
        self.test_site = LocalSite.objects.get(pk=self.test_site.pk)
        self.site_profile = \
            LocalSiteProfile.objects.get(pk=self.site_profile.pk)
        self.site_profile2 = \
            LocalSiteProfile.objects.get(pk=self.site_profile2.pk)
        self.group = Group.objects.get(pk=self.group.pk)

    def _raise_publish_error(self, *args, **kwargs):
        raise NotModifiedError()



"""Diff processing and filtering logic.

.. note::

   Functions in here are considered internal API for the diff processing
   logic. Signatures and behavior may change without a deprecation process.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterable, Iterator, Sequence

    from django.http import HttpRequest
    from typing_extensions import TypeAlias

    from reviewboard.diffviewer.differ import Differ

    _InterdiffChangeRange: TypeAlias = tuple[int, int]


def _get_interdiff_change_ranges(
    *,
    differ: Differ,
) -> Iterator[_InterdiffChangeRange]:
    """Return change ranges for one side of an interdiff.

    This runs a diff on the original and patched file content, collecting
    the ranges of modified lines. These will be used by the interdiff
    filtering algorithm to determine which lines should be shown and which
    should be filtered out.

    The same differ must be used for both range generation and opcode
    generation in order to ensure that ranges will line up.

    The returned ranges are in ``[start, end)`` form, meaning ``start`` is
    inclusive and ``end`` is exclusive (one past the end of the changed
    lines).

    Version Added:
        8.0

    Args:
        differ (reviewboard.diffviewer.differ.Differ):
            The differ used to generate the ranges.

    Yields:
        tuple:
        A 2-tuple of the changed region in the form of:

        Tuple:
            0 (int):
                The start of the range (inclusive).

            1 (int):
                The end of the range (exclusive).
    """
    pending_range: (_InterdiffChangeRange | None) = None

    for tag, i1, i2, j1, j2 in differ.get_opcodes():
        if tag != 'equal' and j1 != j2:
            if pending_range is not None and pending_range[1] >= j1:
                # Extend the existing range.
                pending_range = (pending_range[0], j2)
            else:
                # Yield what we already have before we replace it.
                if pending_range is not None:
                    yield pending_range

                pending_range = (j1, j2)

    # Yield the final range, now that we're done with the opcodes.
    if pending_range is not None:
        yield pending_range


def filter_interdiff_opcodes(
    *,
    opcodes: Iterable[tuple[str, int, int, int, int]],
    differ: Differ,
    filediff_orig_lines: Sequence[str],
    interfilediff_orig_lines: Sequence[str],
    request: (HttpRequest | None) = None,
) -> Iterator[tuple[str, int, int, int, int]]:
    """Filter the opcodes for an interdiff to remove unnecessary lines.

    An interdiff may contain lines of code that have changed as the result of
    updates to the tree between the time that the first and second diff were
    created. This leads to some annoyances when reviewing.

    This function will filter the opcodes to remove as much of this as
    possible. It will only output non-"equal" opcodes if it falls into the
    ranges of lines dictated in the uploaded diff files.

    Version Changed:
        8.0:
        * Support for the v1 and v2 interdiff filtering algorithms have been
          removed. This now uses the v3 algorithm for dual-phase splitting,
          range considerations, and consistent differs, which produces more
          reliable results.

        * Added ``filediff_orig_lines``, ``interfilediff_orig_lines``, and
          ``differ`` parameters, used for the v3 algorithm.

        * Removed the old ``filediff_data`` and ``interfilediff_data``
          parameters.

    Version Changed:
        3.0.18:
        Added the ``request`` argument, and added support for the version 2
        algorithm from Review Board 4.0 (through the :py:data:`~reviewboard
        .diffviewer.features.filter_interdiffs_v2_feature` feature).

    Args:
        opcodes (list of tuple):
            The list of opcodes to filter.

        differ (reviewboard.diffviewer.differ.Differ, optional):
            The differ used for the interdiff comparison.

            This is used to retrieve opcodes for the patched versions of
            both ends of the interdiff.

            Version Added:
                8.0

        filediff_orig_lines (list of str, optional):
            The lines of the original file before the filediff was applied.

            Version Added:
                8.0

        interfilediff_orig_lines (list of str, optional):
            The lines of the original file before the interfilediff was
            applied.

            Version Added:
                8.0

        request (django.http.HttpRequest, optional):
            The HTTP request from the client.

    Yields:
        tuple:
        An opcode to render for the diff.
    """
    # Fetch the changed regions from within the two diff files, and grab the
    # first entry in each, if available.
    differ_cls = type(differ)
    orig_ranges = _get_interdiff_change_ranges(
        differ=differ_cls(
            a=filediff_orig_lines,
            b=differ.a,
            ignore_space=differ.ignore_space,
            compat_version=differ.compat_version,
        ),
    )
    new_ranges = _get_interdiff_change_ranges(
        differ=differ_cls(
            a=interfilediff_orig_lines,
            b=differ.b,
            ignore_space=differ.ignore_space,
            compat_version=differ.compat_version,
        ),
    )

    try:
        orig_range = next(orig_ranges)
    except StopIteration:
        orig_range = None

    try:
        new_range = next(new_ranges)
    except StopIteration:
        new_range = None

    if not orig_range and not new_range:
        # There's nothing in here. Just yield the original opcodes, disabling
        # filtering.
        yield from opcodes

        return

    # Interdiff filtering version 3.
    #
    # Okay, let's go over the approach.
    #
    # There are two main pieces of data we're working with:
    #
    # 1. The diff opcodes coming from the opcode generator.
    #
    #    These featuring a tag (delete, equal, insert, replace), and
    #    [i1, i2), [j1, j2) opcode ranges.
    #
    # 2. Changed ranges for both sides of the interdiff.
    #
    #    To generate each list of changed ranges, the differ (same one used
    #    for this interdiff, with the same settings) is run against the
    #    original and patched versions of the file for the start of the
    #    interdiff and for the end.
    #
    #    The same diff algorithm must be used to ensure that there are no
    #    behavioral differences regarding placement of lines.
    #
    # Based on those, we're going to find which diff opcodes correspond to
    # change ranges, and output those. Any diff opcodes that are outside of
    # those ranges are potential upstream changes not part of the uploaded
    # diff, and will be split out into "filtered-equal" changes. Those will
    # turn back into "equal" changes in post-processing.
    #
    # How we do all this is the complicated part.
    #
    # We start with a main loop that processes the opcodes. This will
    # generally operate on them as they come in from the opcode generation
    # chain, but has a one-entry buffer that can be populated with an opcode
    # that's split and needs further processing.
    #
    # At the start of the loop, we take the change ranges for either side,
    # advancing them if we're past the ends of the previous range.
    #
    # We then begin processing using two phases, each designed to yield an
    # opcode or split it into a "filtered-equal" and a new opcode for
    # further processing.
    #
    # Phase 1: Pre-range split.
    #
    #     An opcode may start before the current range's left boundary
    #     (range[0]) but extend into or past it.
    #
    #     If it starts before a range, the opcode will be split in two at
    #     the start of the range boundary. The part before the range will
    #     normally be emitted as a "filtered-equal" opcode, and the remainder
    #     will be put into the buffer for immediate re-processing. Since this
    #     will start within the range, it will skip phase 1 in the next
    #     iteration and go straight to phase 2.
    #
    #     The exception is when the opcode starts exactly at the exclusive end
    #     of the previous range rather than somewhere before it. That position
    #     is the boundary of a real change, not an upstream change, so the
    #     prefix is emitted as the real opcode tag rather than
    #     "filtered-equal".
    #
    #     If instead the opcode starts within the current range, phase 1
    #     has nothing to do, and we proceed to phase 2.
    #
    #     This is handled for inserts, deletes, and replaces, with their
    #     respective ranges. For inserts and deletes, we only perform the
    #     above checks for the corresponding range. For replaces, we only
    #     split if both sides start before the range, and allow Phase 2 to
    #     handle it if only one side does.
    #
    # Phase 2: Starts-in-range validity check and opcode cap.
    #
    #     An opcode may start within the current range but extend past it.
    #
    #     If it's not even within the range, then it's usually emitted as a
    #     "filtered-equal". Phase 1 would have caught most legitimate cases
    #     of this, so we're normally in a situation where we couldn't even
    #     advance to a range that would contain this, meaning it's likely
    #     past any change ranges in the diff.
    #
    #     The exception is deletes or inserts that the diff algorithm shifted
    #     into the gap between change ranges, which Phase 2 detects separately
    #     (see the boundary validity checks below).
    #
    #     If it starts within the range, but extends past it, then it's split
    #     at the range's right boundary (range[1]). The first part is emitted
    #     as a valid opcode, and the rest is re-queued for further processing
    #     (which may result in further splits from phase 1 or 2, or emitting
    #     as a "filtered-equal").
    #
    #     If it fits fully in the range, then it's emitted as-is.
    #
    #     Like phase 1, this is handled for inserts, deletes, and replaces,
    #     with their respective ranges.
    #
    # The two phases together cover four cases for each opcode:
    #
    # 1. Precedes range (i1 < range[0] and i2 <= range[0]):
    #
    #    The opcode is entirely before the diff change range.
    #
    #    Phase 1 won't kick in (since it's not straddling a range).
    #
    #    Phase 2 will see this as an invalid range and convert to a
    #    "filtered-equal".
    #
    # 2. Straddles start of range (i1 < range[0] and i2 > range[0]):
    #
    #    The opcode straddles the left edge of the range.
    #
    #    Phase 1 will normally emit the part before the range as a
    #    "filtered-equal" and queue processing of the rest. If the opcode
    #    starts exactly at the exclusive end of the previous range rather than
    #    somewhere before it, the prefix sits at a real change boundary and is
    #    emitted as the real opcode tag instead.
    #
    #    Phase 2 will process the rest as one of the next two cases.
    #
    # 3. Fits within range (i1 >= range[0] and i2 <= range[1]):
    #
    #    The opcode fits entirely within the range.
    #
    #    Phase 1 is a no-op (since it doesn't straddle the left of the range).
    #
    #    Phase 2 will emit the opcode as-is.
    #
    # 4. Starts in but exceeds range (i1 >= range[0] and i2 > range[1]):
    #
    #    The opcode starts within the range but extends past it.
    #
    #    Phase 1 is a no-op (since it doesn't straddle the left of the range).
    #
    #    Phase 2 will split at range[1], emit the first part as a valid
    #    opcode, and then either queue processing the rest or emit it as a
    #    "filtered-equal" if there are no more ranges.
    #
    # The result of all this should be "insert", "replace", or "delete"
    # opcodes that correspond to modified ranges within the diff, filtering
    # out anything else as equals.
    #
    # Version 3 of the algorithm was introduced in Review Board 8. Version 2
    # had the Phase 2 behavior, with some workarounds to avoid losing some
    # ranges (but did in some edge cases).
    opcodes_iter = iter(opcodes)
    pending_opcode = None
    prev_orig_range: (_InterdiffChangeRange | None) = None
    prev_new_range: (_InterdiffChangeRange | None) = None
    orig_range_entered: bool = False
    new_range_entered: bool = False
    prev_orig_range_entered: bool = False
    prev_new_range_entered: bool = False

    # Begin the main opcode processing loop.
    while True:
        # First, we'll begin fetching the opcode we want to process, either
        # from the buffer or the generator.
        if pending_opcode is not None:
            # An opcode was put back in the buffer for re-processing. Use
            # that opcode and clear the buffer.
            tag, i1, i2, j1, j2 = pending_opcode
            pending_opcode = None
        else:
            # Fetch the next opcode from the generator.
            try:
                tag, i1, i2, j1, j2 = next(opcodes_iter)
            except StopIteration:
                break

        # Then advance the range on either side if the corresponding diff
        # opcode has moved beyond the previous one.
        #
        # As a reminder, range[1] is exclusive, so range[1] falls outside of
        # the valid ranges for diff opcodes.
        while orig_range is not None and i1 >= orig_range[1]:
            # We've left the range of the current change to consider in the
            # original diff. Move on to the next one.
            prev_orig_range = orig_range
            prev_orig_range_entered = orig_range_entered
            orig_range_entered = False

            try:
                orig_range = next(orig_ranges)
            except StopIteration:
                # There are no more ranges to consider for the original side.
                orig_range = None

        while new_range is not None and j1 >= new_range[1]:
            # We've left the range of the current change to consider in the
            # new diff. Move on to the next one.
            prev_new_range = new_range
            prev_new_range_entered = new_range_entered
            new_range_entered = False

            try:
                new_range = next(new_ranges)
            except StopIteration:
                # There are no more ranges to consider for the modified side.
                new_range = None

        # Phase 1: Pre-range split.
        #
        # If an opcode starts before the current range's left boundary, but
        # also extends into or past it, we split it here. The prefix (up to
        # range[0]) is emitted as a "filtered-equal". The remainder is queued
        # for further processing, and will be handled in phase 2.
        #
        # Without this step, an opcode straddling range[0] would fail the
        # validity check in phase 2 (since i1 < range[0]), and would end up
        # as a "filtered-equal".
        #
        # If it's an equal, we just skip the phase.
        if tag != 'equal':
            orig_needs_split = (
                orig_range is not None and
                tag in {'replace', 'delete'} and
                i1 < orig_range[0] < i2
            )
            new_needs_split = (
                new_range is not None and
                tag in {'replace', 'insert'} and
                j1 < new_range[0] < j2
            )

            if tag == 'replace':
                # A replace opcode will have equal length on both sides, and
                #
                # If only one side starts before its range, phase 2 would
                # consider it valid and perform a cap, and that process would
                # correctly handle alignment of the opcode.
                #
                # If both sides start before their respective range
                # boundaries, they will need to be split.
                needs_split = orig_needs_split and new_needs_split
            else:
                needs_split = orig_needs_split or new_needs_split

            if needs_split:
                if tag == 'insert':
                    assert new_range is not None

                    # Inserts don't affect the orig range, so only the
                    # modified side needs to advance.
                    split_i1 = i1
                    split_j1 = new_range[0]
                elif tag == 'delete':
                    assert orig_range is not None

                    # Deletes don't affect the modified range, so only the
                    # orig side needs to advance.
                    split_i1 = orig_range[0]
                    split_j1 = j1
                else:
                    assert tag == 'replace'

                    # Replace lines affect both ranges, so both need to
                    # advance to a consistent split point.
                    #
                    # We'll compute deltas between the start of the change
                    # range and the opcode position within it.
                    #
                    # The larger of the two deltas will be used for the split
                    # point advancement. This will prevent either side from
                    # ending up behind its range boundary after splitting.
                    if orig_range is not None and i1 < orig_range[0]:
                        orig_delta = orig_range[0] - i1
                    else:
                        orig_delta = 0

                    if new_range is not None and j1 < new_range[0]:
                        new_delta = new_range[0] - j1
                    else:
                        new_delta = 0

                    delta = max(orig_delta, new_delta)
                    split_i1 = i1 + delta
                    split_j1 = j1 + delta

                # Only split if at least one opcode range's split point is
                # before the end of the range.
                if split_i1 < i2 or split_j1 < j2:
                    # Normally the prefix (up to the split point) is from
                    # upstream changes and should be emitted as a
                    # "filtered-equal".
                    #
                    # However, at times the diff algorithm for the ranges and
                    # for the interdiff will calculate some lines differently,
                    # particularly when reordering a series of lines.
                    #
                    # The op can end up being shifted +/- 1 line, causing the
                    # op to start at the exclusive end of the previous range.
                    # When this happens, Phase 1 triggers instead of Phase 2,
                    # and we cut it off as a "filtered-equal".
                    #
                    # If we detect this, we'll want to err on the side of
                    # treating this leading part as a valid tag and then
                    # yielding the rest as a filtered-equal.
                    if tag == 'insert':
                        prefix_is_boundary = (
                            prev_new_range is not None and
                            not prev_new_range_entered and
                            j1 == prev_new_range[1] and
                            new_range is not None
                        )
                    elif tag == 'delete':
                        prefix_is_boundary = (
                            prev_orig_range is not None and
                            not prev_orig_range_entered and
                            i1 == prev_orig_range[1] and
                            orig_range is not None
                        )
                    else:
                        prefix_is_boundary = False

                    if prefix_is_boundary:
                        yield_tag = tag
                    else:
                        yield_tag = 'filtered-equal'

                    yield yield_tag, i1, split_i1, j1, split_j1

                    # Queue the remainder to be processed next.
                    pending_opcode = (tag, split_i1, i2, split_j1, j2)

                    continue

        # Phase 2: Starts-in-range validity check and opcode cap.
        #
        # At this point the opcode starts at or after range[0] (Phase 1
        # ensured that for straddling opcodes). Check whether it falls inside
        # either the orig or new valid range.
        #
        # If either side of the diff opcode has a zero length range
        # (i1 == i2 or j1 == j2), then that's either the new side of a
        # delete or the orig side of an insert. Those aren't considered
        # valid ranges, since there's no content there.
        #
        # Change ranges are [start, end), meaning the end is exclusive. That
        # end may fall within the lines of context or, theoretically, in
        # another range. We'll check if the op is safely within the allowed
        # part of range, for each side.
        orig_starts_valid = (
            i1 != i2 and
            orig_range is not None and
            orig_range[0] <= i1 < orig_range[1]
        )
        new_starts_valid = (
            j1 != j2 and
            new_range is not None and
            new_range[0] <= j1 < new_range[1]
        )

        valid_chunk = orig_starts_valid or new_starts_valid

        if orig_starts_valid and tag != 'equal':
            orig_range_entered = True

        if new_starts_valid and tag != 'equal':
            new_range_entered = True

        if not valid_chunk:
            # The opcode is entirely outside all current valid ranges.
            #
            # Normally we'd emit this as a "filtered-equal", but we can
            # end up in a situation where the diff algorithm for the ranges
            # and for the interdiff will calculate some lines differently,
            # particularly when reordering a series of lines.
            #
            # A delete/insert can end up being shifted past a range's
            # boundary (the exclusive end of the range) by any number of
            # "equal" lines. When advancing the range, these can be missed,
            # causing them to be filtered out when it fails the range
            # validity test.
            #
            # We make a best effort at detecting this by checking if the
            # op starts right at the boundary. For deletes, we can also
            # detect a shift of more than one line by checking whether the
            # start of the delete's range on the "b" side is between two
            # change ranges. Upstream changes shouldn't show up there, so
            # it's a good indicator that it should be included in the
            # interdiff.
            #
            # We can't do the same heuristic for inserts. An insert may be
            # there from upstream changes. All we can safely do is check
            # at the boundary.
            #
            # We factor in whether the previous range was entered (an op
            # starts there) before allowing these heuristics in. That change
            # range may have been split off from a prior range, so we know
            # it's one modified by a diff revision but not matched up
            # against the interdiff, which is why it's considered at all.
            #
            # A subsequent range must exist to rule out actual upstream
            # changes after the last range, and the boundary range cannot
            # have been previously entered in order to rule out ops that
            # start where a previous sequence ended.
            #
            # These heuristics should prevent loss of changes in these
            # situations. There is a potential for upstream changes to end
            # up in the results, but those would be minor and should not
            # impede the review process.
            orig_boundary_valid = (
                tag == 'delete' and
                i1 != i2 and
                orig_range is not None and
                prev_orig_range is not None and
                not prev_orig_range_entered and
                (i1 == prev_orig_range[1] or
                 (i1 > prev_orig_range[1] and
                  prev_new_range is not None and
                  new_range is not None and
                  prev_new_range[1] < j1 < new_range[0]))
            )
            new_boundary_valid = (
                tag == 'insert' and
                j1 != j2 and
                new_range is not None and
                prev_new_range is not None and
                not prev_new_range_entered and
                j1 == prev_new_range[1]
            )

            if orig_boundary_valid or new_boundary_valid:
                # Cap at the start of the next range to avoid spilling into
                # territory that is genuinely upstream.
                if orig_boundary_valid:
                    assert orig_range is not None

                    cap_i2 = orig_range[0]
                else:
                    cap_i2 = i2

                if new_boundary_valid:
                    assert new_range is not None

                    cap_j2 = new_range[0]
                else:
                    cap_j2 = j2

                valid_i2 = min(i2, cap_i2)
                valid_j2 = min(j2, cap_j2)

                yield tag, i1, valid_i2, j1, valid_j2

                # See if we need to split this out for further processing.
                if valid_i2 != i2 or valid_j2 != j2:
                    if orig_boundary_valid:
                        i1 = valid_i2

                    if new_boundary_valid:
                        j1 = valid_j2

                    # Queue the remainder to be processed next.
                    pending_opcode = (tag, i1, i2, j1, j2)

                continue

            # The opcode is genuinely outside all change ranges. Emit it as
            # "filtered-equal". These will get turned back into "equal" chunks
            # in the post-processing step.
            yield 'filtered-equal', i1, i2, j1, j2

            continue

        # The opcode is valid. It may extend past the current range's right
        # boundary (range[1]), so clip each side to the range end.
        if orig_range:
            cap_i2 = orig_range[1]
        else:
            cap_i2 = i2

        if new_range:
            cap_j2 = new_range[1]
        else:
            cap_j2 = j2

        # Determine the end of the valid range for each side.
        if orig_starts_valid:
            valid_i2 = min(i2, cap_i2)
        else:
            valid_i2 = i2

        if new_starts_valid:
            valid_j2 = min(j2, cap_j2)
        else:
            valid_j2 = j2

        if tag in {'equal', 'replace'}:
            # A replace (or equal) opcode must always be of equal length on
            # both the orig and modified sides. We need to determine the end
            # point of this capped range:
            #
            # * If the orig side went out of bounds (valid_i2 > cap_i2),
            #   the modified side is in bounds, so we'll use the modified
            #   length.
            #
            # * If the modified side went out of bounds, the orig side wins
            #   (for the same reasons as above).
            #
            # * If both are in bounds, the longer of the two wins. This
            #   prevents discarding lines that one side still considers valid.
            i_diff = valid_i2 - i1
            j_diff = valid_j2 - j1

            if valid_i2 > cap_i2:
                # Sanity-check that valid_j2 is in bounds. We don't need
                # to check this in the following conditionals, though,
                # since that's covered by the conditionals themselves.
                assert valid_j2 <= cap_j2

                max_cap = j_diff
            elif valid_j2 > cap_j2:
                max_cap = i_diff
            else:
                max_cap = max(i_diff, j_diff)

            # Clamp to the actual opcode bounds so valid_i2/valid_j2 cannot
            # exceed i2/j2 (which would spill over into the next opcode).
            max_cap = min(max_cap, i2 - i1, j2 - j1)

            # Set each valid range to be the same length.
            valid_i2 = i1 + max_cap
            valid_j2 = j1 + max_cap

            # Update the caps to match the computed ends of the opcode ranges
            # so that any splitting will start here later.
            cap_i2 = valid_i2
            cap_j2 = valid_j2

        # Emit the valid part of this opcode.
        yield tag, i1, valid_i2, j1, valid_j2

        # Check if we've handled the full opcode (meaning it's fully in
        # range), or if we need to split out part for re-processing.
        if valid_i2 != i2 or valid_j2 != j2:
            # The opcode extended past the current range. Advance i1/j1 to
            # the start of the remaining part of the opcode where it was
            # split.
            if orig_range is not None and i2 + 1 > cap_i2:
                i1 = cap_i2
            elif orig_range is None and valid_i2 < i2:
                i1 = valid_i2

            if new_range is not None and j2 + 1 > cap_j2:
                j1 = cap_j2
            elif new_range is None and valid_j2 < j2:
                j1 = valid_j2

            # Re-queue the split-off opcode so it can be checked in the next
            # iteration.
            pending_opcode = (tag, i1, i2, j1, j2)


def post_process_filtered_equals(opcodes):
    """Post-processes filtered-equal and equal chunks from interdiffs.

    Any filtered-out "filtered-equal" chunks will get turned back into "equal"
    chunks and merged into any prior equal chunks. Likewise, simple "equal"
    chunks will also get merged.

    "equal" chunks that have any indentation information will remain
    their own chunks, with nothing merged in.
    """
    cur_chunk = None

    for tag, i1, i2, j1, j2, meta in opcodes:
        if ((tag == 'equal' and not meta.get('indentation_changes')) or
            tag == 'filtered-equal'):
            # We either have a plain equal chunk without any indentation
            # changes, or a filtered-equal chunk. In these cases, we can
            # safely merge the chunks together and transform them into
            # an "equal" chunk.
            if cur_chunk:
                i1 = cur_chunk[1]
                j1 = cur_chunk[3]
                meta = cur_chunk[5]

            cur_chunk = ('equal', i1, i2, j1, j2, meta)
        else:
            # This is some sort of changed chunk (insert, delete, replace,
            # or equal with indentation changes). Yield the previous chunk
            # we were working with, if any, and then yield the current chunk.
            if cur_chunk:
                yield cur_chunk
                cur_chunk = None

            yield tag, i1, i2, j1, j2, meta

    if cur_chunk:
        yield cur_chunk



"""LegacyFileDiffData model defitnition."""

from django.db import models
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import Base64Field, JSONField


class LegacyFileDiffData(models.Model):
    """Deprecated, legacy class for base64-encoded diff data.

    This is no longer populated, and exists solely to store legacy data
    that has not been migrated to :py:class:`RawFileDiffData`.
    """

    binary_hash = models.CharField(_('hash'), max_length=40, primary_key=True)
    binary = Base64Field(_('base64'))

    extra_data = JSONField(null=True)

    class Meta:
        app_label = 'diffviewer'
        db_table = 'diffviewer_filediffdata'
        verbose_name = _('Legacy File Diff Data')
        verbose_name_plural = _('Legacy File Diff Data Blobs')



"""Unit tests for reviewboard.reviews.views.ReviewsDiffViewerView."""

from __future__ import annotations

import kgb

from reviewboard.attachments.models import FileAttachment
from reviewboard.reviews.ui.base import ReviewUI
from reviewboard.site.urlresolvers import local_site_reverse
from reviewboard.testing.testcase import BaseFileDiffAncestorTests, TestCase


class ReviewsDiffViewerViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for reviewboard.reviews.views.ReviewsDiffViewerView."""

    fixtures = ['test_users', 'test_scmtools']

    # Bug 892
    def test_interdiff(self) -> None:
        """Testing ReviewsDiffViewerView with interdiffs"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request, revision=1)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/readme',
            dest_file='/readme',
            source_revision='d6613f5',
            dest_detail='5b50866',
            diff=(
                b'diff --git a/readme b/readme\n'
                b'index d6613f5..5b50866 100644\n'
                b'--- a/readme\n'
                b'+++ b/readme\n'
                b'@@ -1 +1,3 @@\n'
                b' Hello there\n'
                b'+\n'
                b'+Oh hi!\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/newfile',
            dest_file='/newfile',
            source_revision='PRE-CREATION',
            dest_detail='',
            diff=(
                b'diff --git a/new_file b/new_file\n'
                b'new file mode 100644\n'
                b'index 0000000..ac30bd3\n'
                b'--- /dev/null\n'
                b'+++ b/new_file\n'
                b'@@ -0,0 +1 @@\n'
                b'+This is a new file!\n'
            ))

        diffset = self.create_diffset(review_request, revision=2)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/readme',
            dest_file='/readme',
            source_revision='d6613f5',
            dest_detail='5b50867',
            diff=(
                b'diff --git a/readme b/readme\n'
                b'index d6613f5..5b50867 100644\n'
                b'--- a/readme\n'
                b'+++ b/readme\n'
                b'@@ -1 +1,3 @@\n'
                b' Hello there\n'
                b'+----------\n'
                b'+Oh hi!\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/newfile',
            dest_file='/newfile',
            source_revision='PRE-CREATION',
            dest_detail='',
            diff=(
                b'diff --git a/new_file b/new_file\n'
                b'new file mode 100644\n'
                b'index 0000000..ac30bd4\n'
                b'--- /dev/null\n'
                b'+++ b/new_file\n'
                b'@@ -0,0 +1 @@\n'
                b'+This is a diffent version of this new file!\n'
            ))

        response = self.client.get('/r/1/diff/1-2/')

        # Useful for debugging any actual errors here.
        if response.status_code != 200:
            print('Error: %s' % response.context['error'])
            print(response.context['trace'])

        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context['diff_context']['num_diffs'], 2)

        files = response.context['files']
        self.assertTrue(files)
        self.assertEqual(len(files), 2)

        self.assertEqual(files[0]['orig_filename'], '/newfile')
        self.assertIn('interfilediff', files[0])

        self.assertEqual(files[1]['orig_filename'], '/readme')
        self.assertIn('interfilediff', files[1])

    # Bug 847
    def test_interdiff_new_file(self) -> None:
        """Testing ReviewsDiffViewrView with interdiffs containing new files"""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request, revision=1)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))

        diffset = self.create_diffset(review_request, revision=2)
        self.create_filediff(
            diffset,
            source_file='/diffutils.py',
            dest_file='/diffutils.py',
            source_revision='6bba278',
            dest_detail='465d217',
            diff=(
                b'diff --git a/diffutils.py b/diffutils.py\n'
                b'index 6bba278..465d217 100644\n'
                b'--- a/diffutils.py\n'
                b'+++ b/diffutils.py\n'
                b'@@ -1,3 +1,4 @@\n'
                b'+# diffutils.py\n'
                b' import fnmatch\n'
                b' import os\n'
                b' import re\n'
            ))
        self.create_filediff(
            diffset,
            source_file='/newfile',
            dest_file='/newfile',
            source_revision='PRE-CREATION',
            dest_detail='',
            diff=(
                b'diff --git a/new_file b/new_file\n'
                b'new file mode 100644\n'
                b'index 0000000..ac30bd4\n'
                b'--- /dev/null\n'
                b'+++ b/new_file\n'
                b'@@ -0,0 +1 @@\n'
                b'+This is a diffent version of this new file!\n'
            ))

        response = self.client.get('/r/1/diff/1-2/')

        # Useful for debugging any actual errors here.
        if response.status_code != 200:
            print('Error: %s' % response.context['error'])
            print(response.context['trace'])

        self.assertEqual(response.status_code, 200)

        self.assertEqual(response.context['diff_context']['num_diffs'], 2)

        files = response.context['files']
        self.assertTrue(files)
        self.assertEqual(len(files), 1)

        self.assertEqual(files[0]['orig_filename'], '/newfile')
        self.assertIn('interfilediff', files[0])

    def test_with_filenames_option(self) -> None:
        """Testing ReviewsDiffViewerView with ?filenames=..."""
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff1 = self.create_filediff(diffset,
                                         source_file='src/main/test.c',
                                         dest_file='src/main/test.cpp')
        filediff2 = self.create_filediff(diffset,
                                         source_file='docs/README.txt',
                                         dest_file='docs/README2.txt')
        filediff3 = self.create_filediff(diffset,
                                         source_file='test.txt',
                                         dest_file='test.rst')
        filediff4 = self.create_filediff(diffset,
                                         source_file='/lib/lib.h',
                                         dest_file='/lib/lib.h')
        self.create_filediff(diffset,
                             source_file='unmatched',
                             dest_file='unmatched')

        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }),
            {
                'filenames': '*/test.cpp,*.txt,/lib/*',
            })
        self.assertEqual(response.status_code, 200)

        files = response.context['files']
        self.assertEqual({file_info['filediff'] for file_info in files},
                         {filediff1, filediff2, filediff3, filediff4})

    def test_with_filenames_option_normalized(self) -> None:
        """Testing ReviewsDiffViewerView with ?filenames=... values normalized
        """
        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff1 = self.create_filediff(diffset,
                                         source_file='src/main/test.c',
                                         dest_file='src/main/test.cpp')
        filediff2 = self.create_filediff(diffset,
                                         source_file='docs/README.txt',
                                         dest_file='docs/README2.txt')
        filediff3 = self.create_filediff(diffset,
                                         source_file='test.txt',
                                         dest_file='test.rst')
        filediff4 = self.create_filediff(diffset,
                                         source_file='/lib/lib.h',
                                         dest_file='/lib/lib.h')
        self.create_filediff(diffset,
                             source_file='unmatched',
                             dest_file='unmatched')

        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }),
            {
                'filenames': ' ,  , */test.cpp,,,*.txt,/lib/*  ',
            })
        self.assertEqual(response.status_code, 200)

        files = response.context['files']
        self.assertEqual({file_info['filediff'] for file_info in files},
                         {filediff1, filediff2, filediff3, filediff4})

    def test_with_static_media(self) -> None:
        """Testing ReviewsDiffViewerView with binary files that have static
        media defined in their review UI
        """
        class DummyReviewUI(ReviewUI):
            js_bundle_names = [
                'js_bundle1',
                'js_bundle2',
            ]

            css_bundle_names = [
                'css_bundle1',
            ]

        self.spy_on(ReviewUI.for_object, op=kgb.SpyOpReturn(DummyReviewUI))

        review_request = self.create_review_request(create_repository=True,
                                                    publish=True)
        diffset = self.create_diffset(review_request)
        filediff1 = self.create_filediff(
            diffset=diffset,
            source_file='foo.png',
            source_revision='123',
            dest_file='foo.png',
            diff=b'diff1',
            binary=True)
        FileAttachment.objects.create_from_filediff(
            filediff1,
            orig_filename='foo.png',
            mimetype='image/png')
        filediff2 = self.create_filediff(diffset,
                                         source_file='src/main/test.c',
                                         dest_file='src/main/test.cpp')

        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': review_request.display_id,
                    'revision': diffset.revision,
                }))
        self.assertEqual(response.status_code, 200)

        files = response.context['diff_context']['files']

        file1 = files[0]
        self.assertEqual(file1['filediff'], {
            'id': filediff1.pk,
            'revision': filediff1.diffset.revision,
        })
        self.assertEqual(file1['js_media'], {'js_bundle1', 'js_bundle2'})
        self.assertEqual(file1['css_media'], {'css_bundle1'})

        file2 = files[1]
        self.assertEqual(file2['filediff'], {
            'id': filediff2.pk,
            'revision': filediff2.diffset.revision,
        })
        self.assertNotIn('js_media', file2)
        self.assertNotIn('cs_media', file2)


class CommitsTests(BaseFileDiffAncestorTests):
    """Tests for ReviewsDiffViewerView with commit history."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.set_up_filediffs()

        review = self.create_review(review_request=self.review_request,
                                    publish=True)
        self.cumulative_comment = self.create_diff_comment(
            review=review,
            filediff=self.diffset.cumulative_files[0])

        commit1 = self.diff_commits[1]
        commit2 = self.diff_commits[2]

        # Comment from the base commit to a tip
        self.commit_comment1 = self.create_diff_comment(
            review=review,
            filediff=commit2.files.get(dest_file='bar'))

        # Comment from one commit to another
        self.commit_comment2 = self.create_diff_comment(
            review=review,
            filediff=commit2.files.get(dest_file='bar'))
        self.commit_comment2.base_filediff_id = \
            commit1.files.get(dest_file='bar').pk
        self.commit_comment2.save(update_fields=['extra_data'])

    def test_comment_data_with_cumulative_diff(self) -> None:
        """Testing ReviewsDiffViewerView comment data with cumulative diff"""
        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': self.review_request.display_id,
                    'revision': self.diffset.revision,
                }))
        self.assertEqual(response.status_code, 200)

        files = response.context['diff_context']['files']

        for f in files:
            comment_blocks = f['serialized_comment_blocks']

            if f['id'] == self.cumulative_comment.filediff_id:
                self.assertEqual(len(comment_blocks), 1)

                comments = list(comment_blocks.values())[0]
                self.assertEqual(comments[0]['comment_id'],
                                 self.cumulative_comment.pk)
            else:
                # All other files in the diff should show no comments on them.
                self.assertEqual(comment_blocks, {})

    def test_comment_data_with_commit_range1(self) -> None:
        """Testing ReviewsDiffViewerView comment data with a commit range from
        the base commit
        """
        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': self.review_request.display_id,
                    'revision': self.diffset.revision,
                }),
            {
                'tip-commit-id': self.diff_commits[2].pk,
            })
        self.assertEqual(response.status_code, 200)

        files = response.context['diff_context']['files']

        for f in files:
            comment_blocks = f['serialized_comment_blocks']

            if f['id'] == self.commit_comment1.filediff_id:
                self.assertEqual(len(comment_blocks), 1)

                comments = list(comment_blocks.values())[0]
                self.assertEqual(comments[0]['comment_id'],
                                 self.commit_comment1.pk)
            else:
                # All other files in the diff should show no comments on them.
                self.assertEqual(comment_blocks, {})

    def test_comment_data_with_commit_range2(self) -> None:
        """Testing ReviewsDiffViewerView comment data with a commit range in
        the middle
        """
        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': self.review_request.display_id,
                    'revision': self.diffset.revision,
                }),
            {
                'base-commit-id': self.diff_commits[1].pk,
                'tip-commit-id': self.diff_commits[2].pk,
            })
        self.assertEqual(response.status_code, 200)

        files = response.context['diff_context']['files']

        for f in files:
            comment_blocks = f['serialized_comment_blocks']

            if f['id'] == self.commit_comment2.filediff_id:
                self.assertEqual(len(comment_blocks), 1)

                comments = list(comment_blocks.values())[0]
                self.assertEqual(comments[0]['comment_id'],
                                 self.commit_comment2.pk)
            else:
                # All other files in the diff should show no comments on them.
                self.assertEqual(comment_blocks, {})

    def test_commits_in_diff_context(self) -> None:
        """Testing that commits are listed properly in diff context"""
        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': self.review_request.display_id,
                    'revision': self.diffset.revision,
                },
            ),
            {
                'base-commit-id': self.diff_commits[1].pk,
                'tip-commit-id': self.diff_commits[2].pk,
            })
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            response.context['diff_context']['commits'],
            [
                {
                    'author_name': 'Author',
                    'commit_id': 'r1',
                    'commit_message': 'Commit message',
                    'id': 1,
                    'parent_id': 'r0',
                },
                {
                    'author_name': 'Author',
                    'commit_id': 'r2',
                    'commit_message': 'Commit message',
                    'id': 2,
                    'parent_id': 'r1',
                },
                {
                    'author_name': 'Author',
                    'commit_id': 'r3',
                    'commit_message': 'Commit message',
                    'id': 3,
                    'parent_id': 'r2',
                },
                {
                    'author_name': 'Author',
                    'commit_id': 'r4',
                    'commit_message': 'Commit message',
                    'id': 4,
                    'parent_id': 'r3',
                },
            ])

    def test_commits_in_diff_context_with_draft(self) -> None:
        """Testing that commits are listed properly in diff context with a
        draft diffset
        """
        self.client.login(username='doc', password='doc')

        draft_diffset = self.create_diffset(
            review_request=self.review_request,
            revision=2,
            draft=True)

        self.create_diffcommit(
            diffset=draft_diffset,
            commit_id='r5',
            parent_id='r4',
            diff_contents=self._COMMITS[0]['diff'])
        draft_diffset.finalize_commit_series(
            cumulative_diff=self._COMMITS[0]['diff'],
            validation_info=None,
            validate=False,
            save=True)

        response = self.client.get(
            local_site_reverse(
                'view-diff-revision',
                kwargs={
                    'review_request_id': self.review_request.display_id,
                    'revision': draft_diffset.revision,
                },
            ))
        self.assertEqual(response.status_code, 200)

        self.assertEqual(
            response.context['diff_context']['commits'],
            [
                {
                    'author_name': 'Author',
                    'commit_id': 'r5',
                    'commit_message': 'Commit message',
                    'id': 5,
                    'parent_id': 'r4',
                },
            ])



"""Unit tests for reviewboard.datagrids.columns.DiffUpdatedSinceColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils import timezone
from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import DiffUpdatedSinceColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class DiffUpdatedSinceColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.DiffUpdatedSinceColumn.

    Version Added:
        8.0
    """

    column = DiffUpdatedSinceColumn()
    fixtures = ['test_users', 'test_scmtools']

    def test_render_data_with_diff(self) -> None:
        """Testing DiffUpdatedSinceColumn.render_data with a diff"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        self.create_diffset(review_request)

        # This is normally set by augmenting the queryset when the diff is
        # added to the history.
        review_request.diffset_history.last_diff_updated = timezone.now()

        result = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(result, SafeString)
        self.assertIn('<time class="timesince"', result)

    def test_render_data_without_diff(self) -> None:
        """Testing DiffUpdatedSinceColumn.render_data without a diff"""
        review_request = self.create_review_request(publish=True)

        # This is normally populated when a diff is added.
        review_request.diffset_history.last_diff_updated = None

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_to_json_with_diff(self) -> None:
        """Testing DiffUpdatedSinceColumn.to_json with a diff"""
        review_request = self.create_review_request(
            create_repository=True,
            publish=True)
        self.create_diffset(review_request)

        # This is normally set by augmenting the queryset when the diff is
        # added to the history.
        review_request.diffset_history.last_diff_updated = timezone.now()

        result = self.column.to_json(self.stateful_column, review_request)

        self.assertIsNotNone(result)

    def test_to_json_without_diff(self) -> None:
        """Testing DiffUpdatedSinceColumn.to_json without a diff"""
        review_request = self.create_review_request(publish=True)

        # This is normally populated when a diff is added.
        review_request.diffset_history.last_diff_updated = None

        self.assertIsNone(
            self.column.to_json(self.stateful_column, review_request))



"""Unit tests for reviewboard.datagrids.columns.BugsColumn.

Version Added:
    8.0
"""

from __future__ import annotations

from django.utils.safestring import SafeString

from reviewboard.datagrids.columns import BugsColumn
from reviewboard.datagrids.tests.base import BaseColumnTestCase


class BugsColumnTests(BaseColumnTestCase):
    """Unit tests for reviewboard.datagrids.columns.BugsColumn.

    Version Added:
        8.0
    """

    column = BugsColumn()

    def test_render_data_with_no_bugs(self) -> None:
        """Testing BugsColumn.render_data with no bugs"""
        review_request = self.create_review_request(publish=True)

        value = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(value, SafeString)
        self.assertEqual(value, '')

    def test_render_data_with_one_bug_no_tracker(self) -> None:
        """Testing BugsColumn.render_data with one bug and no bug tracker"""
        review_request = self.create_review_request(
            bugs_closed='1234',
            publish=True)

        html = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            '<span class="bug">1234</span>')

    def test_render_data_with_one_bug_with_tracker(self) -> None:
        """Testing BugsColumn.render_data with one bug and a bug tracker"""
        repository = self.create_repository(
            tool_name='Test',
            bug_tracker='https://bugs.example.com/%s')
        review_request = self.create_review_request(
            repository=repository,
            bugs_closed='1234',
            publish=True)

        html = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            '<a class="bug" href="/r/1/bugs/1234/">1234</a>')

    def test_render_data_with_multiple_bugs(self) -> None:
        """Testing BugsColumn.render_data with multiple bugs"""
        review_request = self.create_review_request(
            bugs_closed='1234, 5678',
            publish=True)

        html = self.column.render_data(self.stateful_column, review_request)

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            '<span class="bug">1234</span>, <span class="bug">5678</span>')

    def test_to_json_with_no_bugs(self) -> None:
        """Testing BugsColumn.to_json with no bugs"""
        review_request = self.create_review_request(publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            [])

    def test_to_json_with_one_bug_no_tracker(self) -> None:
        """Testing BugsColumn.to_json with one bug and no bug tracker"""
        review_request = self.create_review_request(
            bugs_closed='1234',
            publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            [{'id': '1234'}])

    def test_to_json_with_one_bug_with_tracker(self) -> None:
        """Testing BugsColumn.to_json with one bug and a bug tracker"""
        repository = self.create_repository(
            tool_name='Test',
            bug_tracker='https://bugs.example.com/%s')
        review_request = self.create_review_request(
            repository=repository,
            bugs_closed='1234',
            publish=True)

        self.assertEqual(
            self.column.to_json(self.stateful_column, review_request),
            [
                {
                    'id': '1234',
                    'url': '/r/1/bugs/1234/',
                },
            ])



"""Exception classes for accounts."""

from __future__ import annotations

from django.utils.translation import gettext as _


class LoginNotAllowedError(Exception):
    """An error when a user login is not allowed.

    This error is used when a user cannot log in, for example, if the user is
    marked inactive.

    Version Added:
        6.0.3
    """


class UserQueryError(Exception):
    """An error for when a user query fails during user population.

    This error is used by authentication backends implementing the
    :py:meth:`~reviewboard.accounts.backends.base.BaseAuthBackend
    .populate_users` method to report when an error has occurred that should be
    reported back to the caller.
    """

    def __init__(
        self,
        msg: str,
    ) -> None:
        """Initialize the error.

        Args:
            msg (str):
                The error message to display.
        """
        super().__init__(
            _('Error while populating users from the auth backend: {}')
            .format(msg))


class ServiceAccountUserError(Exception):
    """An error claiming, finding, or setting up a service account user.

    Version Added:
        8.1
    """



"""Test case support for reviewboard.certs.

Version Added:
    6.0
"""

from __future__ import annotations

import os
import shutil
import ssl
from datetime import datetime, timedelta
from http.client import HTTPConnection
from ssl import VerifyMode
from typing import TYPE_CHECKING
from urllib.request import HTTPHandler, HTTPSHandler

import kgb
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509.oid import NameOID
from typing_extensions import Final

from reviewboard.certs.manager import CertificateManager, cert_manager
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from collections.abc import Mapping

    TestCaseMixinBase = TestCase
else:
    TestCaseMixinBase = object


TEST_TRUST_CERT_PEM = b"""
Junk area. This is a trust cert.

-----BEGIN CERTIFICATE-----
MIIEMTCCAxmgAwIBAgIUINE3GHxmshQ/dNCYst/vdY1A1FQwDQYJKoZIhvcNAQEL
BQAwgacxCzAJBgNVBAYTAlVTMRMwEQYDVQQIDApDYWxpZm9ybmlhMRIwEAYDVQQH
DAlQYWxvIEFsdG8xFjAUBgNVBAoMDUJlYW5iYWcsIEluYy4xHTAbBgNVBAsMFFNh
bXBsZSBEb2N1bWVudGF0aW9uMRQwEgYDVQQDDAtleGFtcGxlLmNvbTEiMCAGCSqG
SIb3DQEJARYTc3VwcG9ydEBleGFtcGxlLmNvbTAeFw0yMzA3MTQwNzUwMzBaFw0y
NDA3MTMwNzUwMzBaMIGnMQswCQYDVQQGEwJVUzETMBEGA1UECAwKQ2FsaWZvcm5p
YTESMBAGA1UEBwwJUGFsbyBBbHRvMRYwFAYDVQQKDA1CZWFuYmFnLCBJbmMuMR0w
GwYDVQQLDBRTYW1wbGUgRG9jdW1lbnRhdGlvbjEUMBIGA1UEAwwLZXhhbXBsZS5j
b20xIjAgBgkqhkiG9w0BCQEWE3N1cHBvcnRAZXhhbXBsZS5jb20wggEiMA0GCSqG
SIb3DQEBAQUAA4IBDwAwggEKAoIBAQCx3i72hGP0QiEFjz7D0RUKQdvGsS+8VRan
k/35rMgkfOMNDhOrMUuEViC5RsKUPu8d+8nDKXuKdhWvtIKUHjeXeSvTn7/aDZh9
vt1hxW/lXc9qornncLIgKyJ6rPUeO/Nt5W9nNDTHQO4KOymHmxR48dURT+O2Eluw
rXHcRNSQy3yzT+2KOlHyeUc+rXgKMwbKkm5kD6VSqYBgibcvJ27DZXiPoFRfP7cT
TJC9fcFhsOZKN2H1yYH2yEIMXe6gpxVQ9fEJpv5qlnAs9UneusIXEEfvGomTxpem
M2Y2Vc/KCq0rhBUFPUpQH4yfj79/dxTJPOdKlVapbDpScsPxp8WrAgMBAAGjUzBR
MB0GA1UdDgQWBBSb8YQ1CsPdeenp4A7LCpzyDtC/NTAfBgNVHSMEGDAWgBSb8YQ1
CsPdeenp4A7LCpzyDtC/NTAPBgNVHRMBAf8EBTADAQH/MA0GCSqGSIb3DQEBCwUA
A4IBAQAZ8oXMDH19AKg2uh/4NJtZJA5mRNUCv368R1EiNgw7dqQm6fhkCZCNu1z3
ISEHy2kFeiG+gEDGELOBW/vZrbzPCy0J3I0GobKBlbASEwi21wjQwP0e+tH90h0g
eH1z1ftbAuy06osMFoYcH81+Nb8B9cRVVxgVF/+S2jMUbO3lCAnuyxijH5ShGIt/
fXYd2Fc6Mzk0+krk3+IxluV2O8Pffx8+7zcwUgeDNalNcD6/xjRWkXzD8SuD5nUX
e8K8nyAoCVl6E0HJL7cKdQ9/SkCivQjCO5jK+s8ANOeBNDUSATgQvep9VAk8UWo+
17dcTQ+tCBoxo1eKt5zewpzlYYBV
-----END CERTIFICATE-----

More junk...
"""

TEST_TRUST_SAN_CERT_PEM = b"""
Junk area. This is a trust cert with SAN data.

-----BEGIN CERTIFICATE-----
MIIEXjCCAkagAwIBAgIUH/aJUMZgutxwbM0xXq8DRrfVR+owDQYJKoZIhvcNAQEL
BQAwFzEVMBMGA1UEAwwMVGVzdCBMREFQIENBMB4XDTI2MDMxODAzMjc0OVoXDTM2
MDMxNTAzMjc0OVowGzEZMBcGA1UEAwwQbWFpbi5leGFtcGxlLmNvbTCCASIwDQYJ
KoZIhvcNAQEBBQADggEPADCCAQoCggEBAL0eN9jF/qSqGXKVxu4T9gy+gt2JQYIK
AyemKCs9VZPHTcl+I1pHGGF2wwDDFne9i0JfywSXioLmnkKVar91YlgoCumqJqSQ
5KskzarN5bGy+2V8wU/uhP4sD+gocVxFWPBLAQvJAMwaKtYBtUL7U9Ko9T9BaYDy
pdk+0scFBgfpPYBzcB0UaDIhQ1aM2wd5eih1Gk66N0ZlWwJ4M4Cvtn2L+a10SoXW
ipSFOeeU6LBMTl1vTgsok+4FF/Bf1BqdOTIXRqCXgeq4+v2OnpCLm0G81tyYj1lx
7gnpTvyHpVTR1FUWnFI1zJ3V1RceifgBJoSHYSytfd1maF1eXgDJndECAwEAAaOB
nTCBmjAJBgNVHRMEAjAAMA4GA1UdDwEB/wQEAwIFoDATBgNVHSUEDDAKBggrBgEF
BQcDATAoBgNVHREEITAfghB0ZXN0LmV4YW1wbGUuY29tggtleGFtcGxlLmNvbTAd
BgNVHQ4EFgQURPbKJRFrb6AbUB4GuvEWMIKwCFcwHwYDVR0jBBgwFoAUFZ8R0Mgm
CmxbG0buLdWs5L+1kT0wDQYJKoZIhvcNAQELBQADggIBAA+7n5ZP1fZRtP5mMYl/
U67brXmg9xToesA2vNxyBv9Mb951bqVZKJgGS6i8BkkhhEo90rvv8eIQjfb/owcu
U1hWh+aMXCebrD9T8neTBk+I6lHjbxKFzc3L0uUNmu6fYTqY0hKnsnPgm93V0edd
ObpImOna9vVWtXTsfP0mY9OaaitC4jpaWXSZglTshv0mI23BP3PTCCDn4stY2njt
Oq3j1k+XKjb4p5BTdymDiYh7BdPxIerRbNfHNXwm6YdW30FEdOG5OWj6cHqWNo6n
3+kOAnZ13rasGHgkwZwc5Wzu+U/1uTsjSX35TMbVBIztfZEcTWStqGGbmIp7dXfg
S0esl8aeptVRgcr2t0htw3LkkBmIo6/UBvVbqfJKlqJp5M/sJVgGWZ7rHUk7odhm
8Fw28Hi6R/LD1k5M+SWzQVLZxGMpLObtwRki5lj9FBpSNCctOXh6CwnPmgnkcW9c
ZVqWsFbYzMKGdcDm+azKLR43xHpnbq9ZC+GBfS1dkMRY6Ax13AAHvANBI79hXMJq
pMwwml7ru5CfRaBL/w0VKvaCVza7zRr+DEB9wVTA2JjfLdfMY8raq9jETEcx8n+H
LgIABEl9Djw4VXkTqTSHNbRSKg6IzOK3Ve0rJruJcqPUiCea9c8rwwW3+dBkP3Kr
gynA+/8MseuiQ2HUkoEZOkIy
-----END CERTIFICATE-----

More junk...
"""

TEST_CLIENT_CERT_PEM = b"""
Junk area. This is a client auth cert.

-----BEGIN CERTIFICATE-----
MIIEMTCCAxmgAwIBAgIUINE3GHxmshQ/dNCYst/vdY1A1FQwDQYJKoZIhvcNAQEL
BQAwgacxCzAJBgNVBAYTAlVTMRMwEQYDVQQIDApDYWxpZm9ybmlhMRIwEAYDVQQH
DAlQYWxvIEFsdG8xFjAUBgNVBAoMDUJlYW5iYWcsIEluYy4xHTAbBgNVBAsMFFNh
bXBsZSBEb2N1bWVudGF0aW9uMRQwEgYDVQQDDAtleGFtcGxlLmNvbTEiMCAGCSqG
SIb3DQEJARYTc3VwcG9ydEBleGFtcGxlLmNvbTAeFw0yMzA3MTQwNzUwMzBaFw0y
NDA3MTMwNzUwMzBaMIGnMQswCQYDVQQGEwJVUzETMBEGA1UECAwKQ2FsaWZvcm5p
YTESMBAGA1UEBwwJUGFsbyBBbHRvMRYwFAYDVQQKDA1CZWFuYmFnLCBJbmMuMR0w
GwYDVQQLDBRTYW1wbGUgRG9jdW1lbnRhdGlvbjEUMBIGA1UEAwwLZXhhbXBsZS5j
b20xIjAgBgkqhkiG9w0BCQEWE3N1cHBvcnRAZXhhbXBsZS5jb20wggEiMA0GCSqG
SIb3DQEBAQUAA4IBDwAwggEKAoIBAQCx3i72hGP0QiEFjz7D0RUKQdvGsS+8VRan
k/35rMgkfOMNDhOrMUuEViC5RsKUPu8d+8nDKXuKdhWvtIKUHjeXeSvTn7/aDZh9
vt1hxW/lXc9qornncLIgKyJ6rPUeO/Nt5W9nNDTHQO4KOymHmxR48dURT+O2Eluw
rXHcRNSQy3yzT+2KOlHyeUc+rXgKMwbKkm5kD6VSqYBgibcvJ27DZXiPoFRfP7cT
TJC9fcFhsOZKN2H1yYH2yEIMXe6gpxVQ9fEJpv5qlnAs9UneusIXEEfvGomTxpem
M2Y2Vc/KCq0rhBUFPUpQH4yfj79/dxTJPOdKlVapbDpScsPxp8WrAgMBAAGjUzBR
MB0GA1UdDgQWBBSb8YQ1CsPdeenp4A7LCpzyDtC/NTAfBgNVHSMEGDAWgBSb8YQ1
CsPdeenp4A7LCpzyDtC/NTAPBgNVHRMBAf8EBTADAQH/MA0GCSqGSIb3DQEBCwUA
A4IBAQAZ8oXMDH19AKg2uh/4NJtZJA5mRNUCv368R1EiNgw7dqQm6fhkCZCNu1z3
ISEHy2kFeiG+gEDGELOBW/vZrbzPCy0J3I0GobKBlbASEwi21wjQwP0e+tH90h0g
eH1z1ftbAuy06osMFoYcH81+Nb8B9cRVVxgVF/+S2jMUbO3lCAnuyxijH5ShGIt/
fXYd2Fc6Mzk0+krk3+IxluV2O8Pffx8+7zcwUgeDNalNcD6/xjRWkXzD8SuD5nUX
e8K8nyAoCVl6E0HJL7cKdQ9/SkCivQjCO5jK+s8ANOeBNDUSATgQvep9VAk8UWo+
17dcTQ+tCBoxo1eKt5zewpzlYYBV
-----END CERTIFICATE-----

More junk...
"""

TEST_CLIENT_KEY_PEM = b"""
Junk area. This is a client auth cert key.

-----BEGIN PRIVATE KEY-----
MIIEvAIBADANBgkqhkiG9w0BAQEFAASCBKYwggSiAgEAAoIBAQCx3i72hGP0QiEF
jz7D0RUKQdvGsS+8VRank/35rMgkfOMNDhOrMUuEViC5RsKUPu8d+8nDKXuKdhWv
tIKUHjeXeSvTn7/aDZh9vt1hxW/lXc9qornncLIgKyJ6rPUeO/Nt5W9nNDTHQO4K
OymHmxR48dURT+O2EluwrXHcRNSQy3yzT+2KOlHyeUc+rXgKMwbKkm5kD6VSqYBg
ibcvJ27DZXiPoFRfP7cTTJC9fcFhsOZKN2H1yYH2yEIMXe6gpxVQ9fEJpv5qlnAs
9UneusIXEEfvGomTxpemM2Y2Vc/KCq0rhBUFPUpQH4yfj79/dxTJPOdKlVapbDpS
csPxp8WrAgMBAAECggEABFvTfssDwAqW0JIQEbBf+Z5fimDxMIZNRdIEmUe4p9w/
nCRKKxnMJfQOXTv0rLlWFsAC07uCgYQfR+z+fi63YgjgIBF8HBXVNM+mkSzLby17
VbujHp7OXqdv8t2mLBWAA6NptXe8C1319143yFDukYArnn78r4uHn67AaYtuQhYP
7zsIO5fa0GRLy4HJvOkT4A73P1HLjdM6Sqhaj/mjltivxkFZ1gXmrPHvVJ3mtYfR
qSBPSG/2RJFjbTsY73yw+MYntl+rwlB2Lk7lSOS1Lqqqvg3h4Jefqd5Yq3JwQ5at
XLHMAX8yaNrIHvWhoFa6i4UZh60p0tyRdCoeNKCiEQKBgQC76hDqoocw6PxMwSlC
Pn4aVIGUaYR3pmMpZKjNX8MfVBN778pc9N3kc7KsdChjbofugN0VBbR4wkBun9wN
kniUpvn/TmYkDMuHeVGU0h/ky8SZ9wiTMVyCV/pVQ4Udec5HxAq2pr9G+du0VMTl
95cL5HRbZNKs0ITtku1e67bh/wKBgQDyUEPhAxpbVsfeJelkOtbpF5DNf/Th26br
PwsereKgVXP/+yc/PxENyURogJvuI9paJX3xAMfF2l0bX/2MeqJ5aq5lsSZe/cwa
8zqRDtZ4A5axXGTcze5Yp3yRJA2iXm6dnSrUwhYZAekmHyfY5d2piBFipeOUe8AJ
XRQo+BlEVQKBgFg0NCCOjXqNwe+dM2qevr5JMFGjfcRT23PQhdNwwbvw0Px5v7kG
TykujY2ZMGQLu85dumhltyJ/u5Kxgq70M5lDD1GhURxWzSoX7g023DWe1/eVFvdQ
UiX0edKP0OnXBNZ21LiLaHk+SrxIleuD3eh9IYjMuH7ZmauSLc1CQyOtAoGAO6kX
jWc+Lg8H1uJuvIbgMzPiKza+DH4QcwtqqXsWuXNQxE4KM8BLaqGLfk9sFQ4uzNM+
VwBiL4y8L+lKfDQVnN9vYQpk8C58+oW4fc6xb4sypyigjN0HOjzCptnBoaCui5AO
46OF3VJjHUe+f+DY2sztuuQtTHnLpoKMXCn5zqECgYA2Tfl2W6b1CM2WE19AAOpF
8yN2IJOgVgkkI/dt0qJvr3Kc5er9rvNxlJXFdm+NuteJ7/YffHlffBilYarjTaWV
h+SSpyEeltkqJ5D109FjM2UkendmrGyrnqcusrlqJ/QAyKLTWQFv5VZiWiXy43Ve
9Gp0sLp58Abaa3a/cGt7JQ==
-----END PRIVATE KEY-----

More junk...
"""

TEST_CERT_BUNDLE_PEM = b"""
Here's a cert bundle.

-----BEGIN CERTIFICATE-----
MIIFYDCCBEigAwIBAgIQQAF3ITfU6UK47naqPGQKtzANBgkqhkiG9w0BAQsFADA/
MSQwIgYDVQQKExtEaWdpdGFsIFNpZ25hdHVyZSBUcnVzdCBDby4xFzAVBgNVBAMT
DkRTVCBSb290IENBIFgzMB4XDTIxMDEyMDE5MTQwM1oXDTI0MDkzMDE4MTQwM1ow
TzELMAkGA1UEBhMCVVMxKTAnBgNVBAoTIEludGVybmV0IFNlY3VyaXR5IFJlc2Vh
cmNoIEdyb3VwMRUwEwYDVQQDEwxJU1JHIFJvb3QgWDEwggIiMA0GCSqGSIb3DQEB
AQUAA4ICDwAwggIKAoICAQCt6CRz9BQ385ueK1coHIe+3LffOJCMbjzmV6B493XC
ov71am72AE8o295ohmxEk7axY/0UEmu/H9LqMZshftEzPLpI9d1537O4/xLxIZpL
wYqGcWlKZmZsj348cL+tKSIG8+TA5oCu4kuPt5l+lAOf00eXfJlII1PoOK5PCm+D
LtFJV4yAdLbaL9A4jXsDcCEbdfIwPPqPrt3aY6vrFk/CjhFLfs8L6P+1dy70sntK
4EwSJQxwjQMpoOFTJOwT2e4ZvxCzSow/iaNhUd6shweU9GNx7C7ib1uYgeGJXDR5
bHbvO5BieebbpJovJsXQEOEO3tkQjhb7t/eo98flAgeYjzYIlefiN5YNNnWe+w5y
sR2bvAP5SQXYgd0FtCrWQemsAXaVCg/Y39W9Eh81LygXbNKYwagJZHduRze6zqxZ
Xmidf3LWicUGQSk+WT7dJvUkyRGnWqNMQB9GoZm1pzpRboY7nn1ypxIFeFntPlF4
FQsDj43QLwWyPntKHEtzBRL8xurgUBN8Q5N0s8p0544fAQjQMNRbcTa0B7rBMDBc
SLeCO5imfWCKoqMpgsy6vYMEG6KDA0Gh1gXxG8K28Kh8hjtGqEgqiNx2mna/H2ql
PRmP6zjzZN7IKw0KKP/32+IVQtQi0Cdd4Xn+GOdwiK1O5tmLOsbdJ1Fu/7xk9TND
TwIDAQABo4IBRjCCAUIwDwYDVR0TAQH/BAUwAwEB/zAOBgNVHQ8BAf8EBAMCAQYw
SwYIKwYBBQUHAQEEPzA9MDsGCCsGAQUFBzAChi9odHRwOi8vYXBwcy5pZGVudHJ1
c3QuY29tL3Jvb3RzL2RzdHJvb3RjYXgzLnA3YzAfBgNVHSMEGDAWgBTEp7Gkeyxx
+tvhS5B1/8QVYIWJEDBUBgNVHSAETTBLMAgGBmeBDAECATA/BgsrBgEEAYLfEwEB
ATAwMC4GCCsGAQUFBwIBFiJodHRwOi8vY3BzLnJvb3QteDEubGV0c2VuY3J5cHQu
b3JnMDwGA1UdHwQ1MDMwMaAvoC2GK2h0dHA6Ly9jcmwuaWRlbnRydXN0LmNvbS9E
U1RST09UQ0FYM0NSTC5jcmwwHQYDVR0OBBYEFHm0WeZ7tuXkAXOACIjIGlj26Ztu
MA0GCSqGSIb3DQEBCwUAA4IBAQAKcwBslm7/DlLQrt2M51oGrS+o44+/yQoDFVDC
5WxCu2+b9LRPwkSICHXM6webFGJueN7sJ7o5XPWioW5WlHAQU7G75K/QosMrAdSW
9MUgNTP52GE24HGNtLi1qoJFlcDyqSMo59ahy2cI2qBDLKobkx/J3vWraV0T9VuG
WCLKTVXkcGdtwlfFRjlBz4pYg1htmf5X6DYO8A4jqv2Il9DjXA6USbW1FzXSLr9O
he8Y4IWS6wY7bCkjCWDcRQJMEhg76fsO3txE+FiYruq9RUWhiF1myv4Q6W+CyBFC
Dfvp7OOGAN6dEOM4+qR9sdjoSYKEBpsr6GtPAQw4dy753ec5
-----END CERTIFICATE-----

Another entry:
-----BEGIN CERTIFICATE-----
MIIFFjCCAv6gAwIBAgIRAJErCErPDBinU/bWLiWnX1owDQYJKoZIhvcNAQELBQAw
TzELMAkGA1UEBhMCVVMxKTAnBgNVBAoTIEludGVybmV0IFNlY3VyaXR5IFJlc2Vh
cmNoIEdyb3VwMRUwEwYDVQQDEwxJU1JHIFJvb3QgWDEwHhcNMjAwOTA0MDAwMDAw
WhcNMjUwOTE1MTYwMDAwWjAyMQswCQYDVQQGEwJVUzEWMBQGA1UEChMNTGV0J3Mg
RW5jcnlwdDELMAkGA1UEAxMCUjMwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
AoIBAQC7AhUozPaglNMPEuyNVZLD+ILxmaZ6QoinXSaqtSu5xUyxr45r+XXIo9cP
R5QUVTVXjJ6oojkZ9YI8QqlObvU7wy7bjcCwXPNZOOftz2nwWgsbvsCUJCWH+jdx
sxPnHKzhm+/b5DtFUkWWqcFTzjTIUu61ru2P3mBw4qVUq7ZtDpelQDRrK9O8Zutm
NHz6a4uPVymZ+DAXXbpyb/uBxa3Shlg9F8fnCbvxK/eG3MHacV3URuPMrSXBiLxg
Z3Vms/EY96Jc5lP/Ooi2R6X/ExjqmAl3P51T+c8B5fWmcBcUr2Ok/5mzk53cU6cG
/kiFHaFpriV1uxPMUgP17VGhi9sVAgMBAAGjggEIMIIBBDAOBgNVHQ8BAf8EBAMC
AYYwHQYDVR0lBBYwFAYIKwYBBQUHAwIGCCsGAQUFBwMBMBIGA1UdEwEB/wQIMAYB
Af8CAQAwHQYDVR0OBBYEFBQusxe3WFbLrlAJQOYfr52LFMLGMB8GA1UdIwQYMBaA
FHm0WeZ7tuXkAXOACIjIGlj26ZtuMDIGCCsGAQUFBwEBBCYwJDAiBggrBgEFBQcw
AoYWaHR0cDovL3gxLmkubGVuY3Iub3JnLzAnBgNVHR8EIDAeMBygGqAYhhZodHRw
Oi8veDEuYy5sZW5jci5vcmcvMCIGA1UdIAQbMBkwCAYGZ4EMAQIBMA0GCysGAQQB
gt8TAQEBMA0GCSqGSIb3DQEBCwUAA4ICAQCFyk5HPqP3hUSFvNVneLKYY611TR6W
PTNlclQtgaDqw+34IL9fzLdwALduO/ZelN7kIJ+m74uyA+eitRY8kc607TkC53wl
ikfmZW4/RvTZ8M6UK+5UzhK8jCdLuMGYL6KvzXGRSgi3yLgjewQtCPkIVz6D2QQz
CkcheAmCJ8MqyJu5zlzyZMjAvnnAT45tRAxekrsu94sQ4egdRCnbWSDtY7kh+BIm
lJNXoB1lBMEKIq4QDUOXoRgffuDghje1WrG9ML+Hbisq/yFOGwXD9RiX8F6sw6W4
avAuvDszue5L3sz85K+EC4Y/wFVDNvZo4TYXao6Z0f+lQKc0t8DQYzk1OXVu8rp2
yJMC6alLbBfODALZvYH7n7do1AZls4I9d1P4jnkDrQoxB3UqQ9hVl3LEKQ73xF1O
yK5GhDDX8oVfGKF5u+decIsH4YaTw7mP3GFxJSqv3+0lUFJoi5Lc5da149p90Ids
hCExroL1+7mryIkXPeFM5TgO9r0rvZaBFOvV2z0gp35Z0+L4WPlbuEjN/lxPFin+
HlUjr8gRsI3qfJOQFy/9rKIJR0Y/8Omwt/8oTWgy1mdeHmmjk7j1nYsvC9JSQ6Zv
MldlTTKB3zhThV1+XWYp6rjd5JW1zbVWEkLNxE7GJThEUG3szgBVGP7pSWTUTsqX
nLRbwHOoq7hHwg==
-----END CERTIFICATE-----

And another:
-----BEGIN CERTIFICATE-----
MIIGUDCCBTigAwIBAgISBI+q2t/HklIfG4tA8/oK77gKMA0GCSqGSIb3DQEBCwUA
MDIxCzAJBgNVBAYTAlVTMRYwFAYDVQQKEw1MZXQncyBFbmNyeXB0MQswCQYDVQQD
EwJSMzAeFw0yMzA2MjAwNDUwNDdaFw0yMzA5MTgwNDUwNDZaMBoxGDAWBgNVBAMT
D3Jldmlld2JvYXJkLm9yZzCCAiIwDQYJKoZIhvcNAQEBBQADggIPADCCAgoCggIB
AL6ZFX3tNiWkZhU/2I1tjYqdnZkRX8XQdg4eEe7zikcQftltWmtgdmkBjSE2HaIk
Y89k23LfL5DHgQeEWnZU1My51YpvAqnUm/sQ+wipzmbPJ9IcKy9YJ3gvkwtZmNXE
RvJUN9r5Vl45RepytRGhhF4F4K4eCubvN798XEbgqghj5Z1JtaUk3ewHSRW0gGMU
nOQhBtN7NXsQh3GNu/caNm4yvXHmqoxrop+be/ZYJRhthMFh0rgFEAsu7TeAUHnr
FTseFj0Rj7AhcdAg5LzdkGFhq4heeHNGQgz6M+fBK189tlRFqsl8C4augC0tTE9Z
a/CeBbHdhGAsoRD6DGsuHdvwu+WmQOZbgNqEuwaFpYfuJpRwlf227Hx6gswFbBfz
Dj9YB8QX/1/C4IDvmy58oFwxqAObRJTLWt9VW5zeOITpRFhoRtcZqZMfLLrN841m
c5ybP7ShKoLVP9BWyQ87Y3pDx17xeS6Kh/fB4IEiCMdEGnmUQ7aWfBROQfDD7cGe
dVk+14VLcy/McvLnc9PNWMEEeI2yO/ZCawdGcUWIRkN0LUTQRKYlm+fVgxjhhDbU
LVb8ceH3W7POKhCIp0G7c8V+eXOaSkOBlsQRr8IoD9urhyXH4tuZ46uPb6SAk2bt
cd1MyuU/K8960M0a8G4+WBOVQK2QsvCH9JKn+188zq8/AgMBAAGjggJ2MIICcjAO
BgNVHQ8BAf8EBAMCBaAwHQYDVR0lBBYwFAYIKwYBBQUHAwEGCCsGAQUFBwMCMAwG
A1UdEwEB/wQCMAAwHQYDVR0OBBYEFBY+a+mumhqtoVXIyLDyBpvM0CNoMB8GA1Ud
IwQYMBaAFBQusxe3WFbLrlAJQOYfr52LFMLGMFUGCCsGAQUFBwEBBEkwRzAhBggr
BgEFBQcwAYYVaHR0cDovL3IzLm8ubGVuY3Iub3JnMCIGCCsGAQUFBzAChhZodHRw
Oi8vcjMuaS5sZW5jci5vcmcvMH8GA1UdEQR4MHaCEHJldmlldy1ib2FyZC5uZXSC
EHJldmlldy1ib2FyZC5vcmeCD3Jldmlld2JvYXJkLm9yZ4IUd3d3LnJldmlldy1i
b2FyZC5uZXSCFHd3dy5yZXZpZXctYm9hcmQub3JnghN3d3cucmV2aWV3Ym9hcmQu
b3JnMBMGA1UdIAQMMAowCAYGZ4EMAQIBMIIBBAYKKwYBBAHWeQIEAgSB9QSB8gDw
AHYAejKMVNi3LbYg6jjgUh7phBZwMhOFTTvSK8E6V6NS61IAAAGI11sJQwAABAMA
RzBFAiEAigvbRUqjNAK/OtlO3D2gbKEzGrI6cjNy2Gzu+9MRLaICIHoQN5y/2XvS
Lzw2r6aPyaC12jzYPcreUkzhivIDAmegAHYAtz77JN+cTbp18jnFulj0bF38Qs96
nzXEnh0JgSXttJkAAAGI11sJTwAABAMARzBFAiACLdcZBHQyjxoC66VuL1Ml7teR
i8RWM4ItVxXTVdstbgIhALRj4ubSbynvZlF9coEvzZwRG0NhhVP15y5EcktHEZGW
MA0GCSqGSIb3DQEBCwUAA4IBAQBbPEpF1kIkua+U/u+3zxel/YE4B43nHi8VI8Lx
Pxzz1u9YsL9q/Kv7+MTR/lqxdYm9+Tgy581mq8GVl4vyXZGS6Z4QM+NqbpUYkbKE
Is9ckyoby3lVet00iVIObU8j9zxNpc6bWdusmuSIKwy4dJusI+wRx0m79fXmffUT
HFXdjGoo/7WyKHKpZIrUiQCEaxyg0d/PR0h6L53Uty5lzC8UqqQPVsYKHde41tgd
pYqtPwuigULIIwPjyCH5aqN9gsJL1mj0oaUKwat8GD+xRAexcaaXtHvStlWMzDmD
R/VelCH+wHaDm4EoroX1lKyRi94bOMDXcDVo0kqBt65rYwrK
-----END CERTIFICATE-----"""

TEST_SHA1 = 'F2:35:0F:BB:34:40:84:78:8B:20:1D:40:B1:4A:17:0C:DE:36:2F:D5'
TEST_SHA1_2 = '00:11:22:33:44:55:66:77:88:99:AA:BB:CC:DD:EE:FF:00:11:22:33'

TEST_SHA256 = (
    '79:19:70:AE:A6:1B:EB:BC:35:7C:B8:54:B1:6A:AD:79:FF:F7:28:69:02:5E:C3:'
    '6F:B3:C2:B4:FD:84:66:DF:8F'
)
TEST_SHA256_2 = (
    '00:11:22:33:44:55:66:77:88:99:AA:BB:CC:DD:EE:FF:00:11:22:33:44:55:66:'
    '77:88:99:AA:BB:CC:DD:EE:FF'
)


class MockHTTPResponse:
    """A mock HTTP response for certificate testing.

    This allows control over HTTP response codes, headers, and payload data.

    Version Added:
        8.1
    """

    ######################
    # Instance variables #
    ######################

    #: The data in the response body.
    data: bytes

    #: The HTTP status code.
    code: int

    #: The HTTP status reason.
    reason: str

    #: The normalized headers in the response.
    headers: Mapping[str, str]

    #: The URL returning the response.
    url: str

    def __init__(
        self,
        *,
        data: bytes = b'test',
        code: int = 200,
        reason: str = 'OK',
        headers: (Mapping[str, str] | None) = None,
        url: str = 'https://example.com',
    ) -> None:
        """Initialize the mock HTTP response.

        Args:
            data (bytes, optional):
                The data in the response body.

            code (int, optional):
                The HTTP status code.

                This defaults to 200.

            reason (str, optional):
                The HTTP status reason.

                This defaults to "OK".

            headers (dict, optional):
                The normalized headers in the response.

                This defaults to a :mailheader:`content-type` of
                ``text/plain``.

            url (str, optional):
                The URL returning the response.
        """
        self.data = data
        self.code = code
        self.reason = reason
        self.headers = headers or {
            'content-type': 'text/plain',
        }
        self.url = url

    def getcode(self) -> int:
        """Return the HTTP status code for the response.

        Returns:
            int:
            The HTTP status code.
        """
        return self.code

    def geturl(self) -> str:
        """Return the URL returning the response.

        Returns:
            str:
            The URL.
        """
        return self.url

    def info(self) -> Mapping[str, str]:
        """Return the normalized headers for the response.

        Returns:
            dict:
            The normalized headers.
        """
        return self.headers

    def read(self) -> bytes:
        """Return the response body.

        Returns:
            bytes:
            The response body.
        """
        return self.data

    def close(self) -> None:
        """Close the response."""


class CaptureSSLContext:
    """SSL context that captures loaded state.

    Version Added:
        8.0
    """

    ######################
    # Instance variables #
    ######################

    #: Loaded CA bundle strings.
    cadatas: list[bytes | str | None]

    #: Loaded CA bundle files.
    cafiles: list[str | None]

    #: Loaded CA bundle paths.
    capaths: list[str | None]

    #: Loaded trust or mTLS certificate files.
    certfiles: list[str | None]

    #: Whether hostnames are checked.
    check_hostname: bool

    #: Loaded mTLS key files.
    keyfiles: list[str | None]

    #: Loaded mTLS key passwords.
    passwords: list[str | None]

    def __init__(self) -> None:
        """Initialize the context."""
        self.cadatas = []
        self.cafiles = []
        self.capaths = []
        self.certfiles = []
        self.keyfiles = []
        self.passwords = []
        self.check_hostname = True
        self.verify_mode = VerifyMode.CERT_REQUIRED

    def load_verify_locations(
        self,
        cafile: (str | None) = None,
        capath: (str | None) = None,
        cadata: (bytes | str | None) = None
    ) -> None:
        """Load CA data for verification.

        Args:
            cafile (str, optional):
                The CA file path to load.

            capath (str, optional):
                The CA directory path to load.

            cadata (bytes or str, optional):
                The CA file data to load.
        """
        if cafile:
            self.cafiles.append(cafile)

        if capath:
            self.capaths.append(capath)

        if cadata:
            self.cadatas.append(cadata)

    def load_cert_chain(
        self,
        certfile: str,
        keyfile: (str | None) = None,
        password: (str | None) = None,
    ) -> None:
        """Load a certificate chain.

        Args:
            certfile (str):
                The certificate file path.

            keyfile (str, optional):
                The key file path.

            password (str, optional):
                The key password.
        """
        self.certfiles.append(certfile)
        self.keyfiles.append(keyfile)
        self.passwords.append(password)


class CaptureSSLMixin(TestCaseMixinBase):
    """A mixin for unit tests that need to capture HTTP SSL contexts.

    This takes care of setting up the state needed to capture all SSL
    contexts set up for HTTP requests. It will clear out the stored
    certificates before and after tests, set up a mock HTTP response that
    tests can customize, and serve it up for responses.

    Any SSL contexts set up during the test run will be stored in
    :py:attr:`ssl_contexts` for introspection.

    Subclasses must mix in :py:class:`kgb.SpyAgency`.

    Version Added:
        8.1
    """

    ######################
    # Instance variables #
    ######################

    #: The certificate manager used for the test and storage.
    cert_manager: CertificateManager

    #: The list of captured SSL contexts.
    ssl_contexts: list[CaptureSSLContext]

    #: The mock HTTP response to serve.
    #:
    #: Tests can customize this (but cannot replace the instance) If they
    #: need to return more than one response in order of request, they will
    #: need to unspy :py:meth:`http.client.HTTPConnection.getresponse` and
    #: re-spy on it with the results returned in order.
    mock_http_response: MockHTTPResponse

    def setUp(self) -> None:
        """Set up state for the test.

        This will clear out any existing certificate state and begin spying
        on all HTTP(S) operations needed to control communication and test
        results.
        """
        super().setUp()

        storage_path = cert_manager.storage_backend.storage_path
        self.cert_manager = cert_manager

        # Clear out any stored SSL certificate state before the test.
        if os.path.exists(storage_path):
            shutil.rmtree(storage_path)

        self.mock_http_response = MockHTTPResponse()

        # We only need to spy on HTTPConnection. HTTPSConnection inherits
        # these from that.
        self.spy_on(HTTPConnection.request,
                    owner=HTTPConnection,
                    call_original=False)
        self.spy_on(HTTPConnection.getresponse,
                    owner=HTTPConnection,
                    op=kgb.SpyOpReturn(self.mock_http_response))

        self.spy_on(HTTPHandler.http_open,
                    owner=HTTPHandler)
        self.spy_on(HTTPSHandler.https_open,
                    owner=HTTPSHandler)

        # Watch for any created SSL contexts and store them in the list.
        self.ssl_contexts = []

        @self.spy_for(ssl.create_default_context)
        def _create_default_context(*args, **kwargs):
            ssl_context = CaptureSSLContext()
            self.ssl_contexts.append(ssl_context)

            return ssl_context

    def tearDown(self) -> None:
        """Tear down state for the test.

        This will clear out any existing certificate state after the test and
        clear all stored responses and SSL contexts.
        """
        storage_path = self.cert_manager.storage_backend.storage_path

        if os.path.exists(storage_path):
            shutil.rmtree(storage_path)

        super().tearDown()


class CertificateTestCase(TestCase):
    """Base test case for certificate unit tests.

    Version Aded:
        6.0
    """

    #: Path to the base cert testdata directory.
    base_testdata_dir: Final[str] = os.path.abspath(
        os.path.join(__file__, '..', 'testdata'))

    def build_x509_cert(
        self,
        *,
        subject: str = 'example.com',
        issuer: str = 'example.com',
        not_valid_before_delta: timedelta = -timedelta(days=1),
        not_valid_after_delta: timedelta = timedelta(days=1),
    ) -> x509.Certificate:
        """Return a new Cryptography X.509 certificate object.

        Args:
            subject (str, optional):
                The subject name for the certificate.

            issuer (str, optional):
                The issuer name for the certificate.

            not_valid_before_delta (datetime.timedelta, optional):
                The delta relative to now indicating when the certificate is
                first valid.

            not_valid_after_delta (datetime.timedelta, optional):
                The delta relative to now indicating when the certificate
                expires.

        Returns:
            cryptography.x509.Certificate:
            The resulting certificate.
        """
        private_key = rsa.generate_private_key(public_exponent=65537,
                                               key_size=2048)

        return (
            x509.CertificateBuilder()
            .subject_name(x509.Name([
                x509.NameAttribute(NameOID.COMMON_NAME, subject),
            ]))
            .issuer_name(x509.Name([
                x509.NameAttribute(NameOID.COMMON_NAME, issuer),
            ]))
            .not_valid_before(datetime.utcnow() + not_valid_before_delta)
            .not_valid_after(datetime.utcnow() + not_valid_after_delta)
            .serial_number(x509.random_serial_number())
            .public_key(private_key.public_key())
            .sign(private_key=private_key,
                  algorithm=hashes.SHA256())
        )

    def build_x509_cert_pem(self, **kwargs) -> bytes:
        """Return a new certificate as PEM content.

        Args:
            **kwargs (dict):
                Keyword arguments to pass to :py:meth:`build_x509_cert`.

        Returns:
            bytes:
            The PEM content for the certificate.
        """
        return self.build_x509_cert(**kwargs).public_bytes(Encoding.PEM)



"""Unit tests for reviewboard.attachments.models.UserFileAttachment.

Version Added:
    7.0.3:
    This was split off from :py:mod:`reviewboard.attachments.tests`.
"""

from __future__ import annotations

import kgb
import json
import os
from datetime import datetime
from typing import TYPE_CHECKING

from django.contrib.auth.models import AnonymousUser, User
from reviewboard.attachments.forms import UploadUserFileForm
from reviewboard.attachments.tests.base import BaseFileAttachmentTestCase
from reviewboard.site.models import LocalSite

if TYPE_CHECKING:
    from typing import Any


class UserFileAttachmentTests(kgb.SpyAgency, BaseFileAttachmentTestCase):
    """Unit tests for reviewboard.attachments.models.UserFileAttachment.

    Version Added:
        7.0.3:
        This was split off from :py:mod:`reviewboard.attachments.tests`.
    """

    default_expected_extra_data: dict[str, Any] = {
        BaseFileAttachmentTestCase.DEFINED_LOCAL_SITE_KEY: True,
    }

    def test_user_file_add_file_after_create(self) -> None:
        """Testing user FileAttachment create without initial file and
        adding file through update
        """
        user = User.objects.get(username='doc')

        form = UploadUserFileForm(files={})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertFalse(file_attachment.file)
        self.assertEqual(file_attachment.user, user)
        self.assertEqual(file_attachment.extra_data,
                         self.default_expected_extra_data)

        uploaded_file = self.make_uploaded_file()
        form = UploadUserFileForm(files={
            'path': uploaded_file,
        })
        self.assertTrue(form.is_valid())

        file_attachment = form.update(file_attachment)
        file_attachment.refresh_from_db()

        self.assertTrue(os.path.basename(file_attachment.file.name).endswith(
            '__logo.png'))
        self.assertEqual(file_attachment.mimetype, 'image/png')

    def test_user_file_with_upload_file(self) -> None:
        """Testing user FileAttachment create with initial file"""
        user = User.objects.get(username='doc')
        uploaded_file = self.make_uploaded_file()

        form = UploadUserFileForm(files={
            'path': uploaded_file,
        })
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)

        self.assertEqual(file_attachment.user, user)
        self.assertTrue(os.path.basename(file_attachment.file.name).endswith(
            '__logo.png'))
        self.assertEqual(file_attachment.mimetype, 'image/png')
        self.assertEqual(file_attachment.extra_data,
                         self.default_expected_extra_data)

    def test_user_file_with_extra_data(self) -> None:
        """Testing user FileAttachment create with extra data"""
        class TestObject():
            def to_json(self):
                return {
                    'foo': 'bar'
                }

        user = User.objects.get(username='doc')
        uploaded_file = self.make_uploaded_file()

        form = UploadUserFileForm(
            data={
                'extra_data': {
                    'test_bool': True,
                    'test_date': datetime(2023, 1, 26, 5, 30, 3, 123456),
                    'test_int': 1,
                    'test_list': [1, 2, 3],
                    'test_nested_dict': {
                        'foo': 2,
                        'bar': 'baz',
                    },
                    'test_none': None,
                    'test_obj': TestObject(),
                    'test_str': 'test',
                }
            },
            files={'path': uploaded_file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.user, user)
        self.assertTrue(os.path.basename(file_attachment.file.name).endswith(
            '__logo.png'))
        self.assertEqual(file_attachment.mimetype, 'image/png')
        self.assertEqual(file_attachment.extra_data, {
            'test_bool': True,
            'test_date': '2023-01-26T05:30:03.123',
            'test_int': 1,
            'test_list': [1, 2, 3],
            'test_nested_dict': {
                'foo': 2,
                'bar': 'baz',
            },
            'test_none': None,
            'test_obj': {
                'foo': 'bar',
            },
            'test_str': 'test',
            **self.default_expected_extra_data,
        })

    def test_user_file_with_extra_data_string(self) -> None:
        """Testing user FileAttachment create with extra data passed as a
        JSON string
        """
        user = User.objects.get(username='doc')
        uploaded_file = self.make_uploaded_file()

        form = UploadUserFileForm(
            data={
                'extra_data': json.dumps({
                    'test_bool': True,
                    'test_int': 1,
                    'test_list': [1, 2, 3],
                    'test_nested_dict': {
                        'foo': 2,
                        'bar': 'baz',
                    },
                    'test_none': None,
                    'test_str': 'test',
                })
            },
            files={'path': uploaded_file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data, {
            'test_bool': True,
            'test_int': 1,
            'test_list': [1, 2, 3],
            'test_nested_dict': {
                'foo': 2,
                'bar': 'baz',
            },
            'test_none': None,
            'test_str': 'test',
            **self.default_expected_extra_data,
        })

    def test_user_file_with_extra_data_empties(self) -> None:
        """Testing user FileAttachment create with extra data that contains
        empty values
        """
        user = User.objects.get(username='doc')
        uploaded_file = self.make_uploaded_file()

        form = UploadUserFileForm(
            data={
                'extra_data': {}
            },
            files={'path': uploaded_file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data,
                         self.default_expected_extra_data)

        form = UploadUserFileForm(
            data={
                'extra_data': json.dumps(None)
            },
            files={'path': uploaded_file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data,
                         self.default_expected_extra_data)

        form = UploadUserFileForm(
            data={
                'extra_data': None
            },
            files={'path': uploaded_file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data,
                         self.default_expected_extra_data)

        form = UploadUserFileForm(
            data={
                'extra_data': {
                    'test_list': [],
                    'test_nested_dict': {},
                    'test_none': None,
                    'test_str': '',
                }
            },
            files={'path': uploaded_file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data, {
            'test_list': [],
            'test_nested_dict': {},
            'test_none': None,
            'test_str': '',
            **self.default_expected_extra_data,
        })

    def test_user_file_local_sites(self) -> None:
        """Testing user FileAttachment create with local site"""
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.get(name='local-site-1')

        form = UploadUserFileForm(files={})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user, local_site)

        self.assertEqual(file_attachment.user, user)
        self.assertEqual(file_attachment.local_site, local_site)

    def test_user_file_update_with_extra_data(self) -> None:
        """Testing user FileAttachment update with extra data"""
        class TestObject():
            def to_json(self) -> None:
                return {
                    'foo': 'bar'
                }

        user = User.objects.get(username='doc')
        uploaded_file = self.make_uploaded_file()

        form = UploadUserFileForm(files={'path': uploaded_file})
        self.assertTrue(form.is_valid())

        file_attachment = form.create(user)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data,
                         self.default_expected_extra_data)

        form = UploadUserFileForm(
            data={
                'extra_data': {
                    'test_bool': True,
                    'test_date': datetime(2023, 1, 26, 5, 30, 3, 123456),
                    'test_int': 1,
                    'test_list': [1, 2, 3],
                    'test_nested_dict': {
                        'foo': 2,
                        'bar': 'baz',
                    },
                    'test_none': None,
                    'test_obj': TestObject(),
                    'test_str': 'test',
                }
            }
        )
        self.assertTrue(form.is_valid())

        file_attachment = form.update(file_attachment)
        file_attachment.refresh_from_db()

        self.assertEqual(file_attachment.extra_data, {
            'test_bool': True,
            'test_date': '2023-01-26T05:30:03.123',
            'test_int': 1,
            'test_list': [1, 2, 3],
            'test_nested_dict': {
                'foo': 2,
                'bar': 'baz',
            },
            'test_none': None,
            'test_obj': {
                'foo': 'bar',
            },
            'test_str': 'test',
            **self.default_expected_extra_data,
        })

    def test_user_file_is_accessible_by(self) -> None:
        """Testing user FileAttachment.is_accessible_by"""
        creating_user = User.objects.get(username='doc')
        admin_user = User.objects.get(username='admin')
        same_site_user = User.objects.get(username='dopey')
        different_site_user = User.objects.get(username='grumpy')

        local_site = LocalSite.objects.get(name='local-site-1')
        local_site.users.add(same_site_user)

        form = UploadUserFileForm(files={})
        self.assertTrue(form.is_valid())
        file_attachment = form.create(creating_user, local_site)

        self.assertTrue(file_attachment.is_accessible_by(admin_user))
        self.assertTrue(file_attachment.is_accessible_by(creating_user))
        self.assertFalse(file_attachment.is_accessible_by(AnonymousUser()))
        self.assertFalse(file_attachment.is_accessible_by(same_site_user))
        self.assertFalse(file_attachment.is_accessible_by(different_site_user))

    def test_user_file_is_mutable_by(self) -> None:
        """Testing user FileAttachment.is_mutable_by"""
        creating_user = User.objects.get(username='doc')
        admin_user = User.objects.get(username='admin')
        same_site_user = User.objects.get(username='dopey')
        different_site_user = User.objects.get(username='grumpy')

        local_site = LocalSite.objects.get(name='local-site-1')
        local_site.users.add(same_site_user)

        form = UploadUserFileForm(files={})
        self.assertTrue(form.is_valid())
        file_attachment = form.create(creating_user, local_site)

        self.assertTrue(file_attachment.is_mutable_by(admin_user))
        self.assertTrue(file_attachment.is_mutable_by(creating_user))
        self.assertFalse(file_attachment.is_mutable_by(AnonymousUser()))
        self.assertFalse(file_attachment.is_mutable_by(same_site_user))
        self.assertFalse(file_attachment.is_mutable_by(different_site_user))

    def test_get_absolute_url(self) -> None:
        """Testing user FileAttachment.get_absolute_url"""
        user = User.objects.get(username='doc')

        form = UploadUserFileForm(files={})
        self.assertTrue(form.is_valid())
        file_attachment = form.create(user, None)

        self.assertEqual(
            file_attachment.get_absolute_url(),
            f'http://example.com/users/{user.username}/'
            f'file-attachments/{file_attachment.uuid}/')

    def test_get_absolute_url_with_local_site(self) -> None:
        """Testing user FileAttachment.get_absolute_url with a local site"""
        user = User.objects.get(username='doc')
        local_site = LocalSite.objects.get(name='local-site-1')

        form = UploadUserFileForm(files={})
        self.assertTrue(form.is_valid())
        file_attachment = form.create(user, local_site)

        self.assertEqual(
            file_attachment.get_absolute_url(),
            f'http://example.com/s/{local_site.name}/users/{user.username}/'
            f'file-attachments/{file_attachment.uuid}/')

    def test_get_local_site(self) -> None:
        """Testing FileAttachment.get_local_site returns the stored local site
        """
        expected_local_site = self.get_local_site(self.local_site_name)
        user = User.objects.get(username='doc')
        file_attachment = self.create_user_file_attachment(
            user=user,
            local_site=expected_local_site)

        self.spy_on(file_attachment.get_review_request)

        with self.assertNumQueries(0):
            local_site = file_attachment.get_local_site()

        self.assertEqual(local_site, expected_local_site)
        self.assertIsNotNone(local_site)
        self.assertSpyNotCalled(file_attachment.get_review_request)

    def test_get_local_site_none(self) -> None:
        """Testing FileAttachment.get_local_site with local site set to None"""
        user = User.objects.get(username='doc')
        file_attachment = self.create_user_file_attachment(
            user=user)

        with self.assertNumQueries(0):
            local_site = file_attachment.get_local_site()

        self.assertIsNone(local_site)
